# -*- coding: utf-8 -*-
'''

'''

###########################
### 0. Import libralies ###
###########################


# 0.1. libraries for fundamental data science and data processing

import warnings
import os
import pandas as pd
import numpy as np
from scipy import stats
from copy import deepcopy

from igraph import Graph

from ..utility.hdf5_processing import dump_hdf5, load_hdf5

from .use_r_scripts import (_get_network_score_by_Rscripts_inparallel,
                            _check_R_libraries_installation)
from .network_structure_analysis import (plot_degree_distributions,
                                         plot_score_discributions,
                                         plot_network_entropy_distributions)

from .gene_analysis import (plot_scores_as_rank,
                            plot_score_comparison_2D,
                            plot_score_comparison_2D_with_plotly,
                            plot_score_per_cluster,
                            plot_cartography_scatter_per_cluster,
                            plot_cartography_term)

def load_links(file_path):
    """
    Load links object saved as a hdf5 file.

    Args:
        file_path (str): file path.

    Returns:
        Links: loaded links object.

    """
    links = load_hdf5(filename=file_path, obj_class=Links)

    _update_links_object(links)

    return links

def _update_links_object(links):
    if hasattr(links, "thread_number"): # if the links object has old version attribute
        if links.thread_number is not None:
            links.threshold_number = deepcopy(links.thread_number)
            delattr(links, "thread_number")

class Links():
    """
    This is a class for the processing and visualization of GRNs.
    Links object stores cluster-specific GRNs and metadata.
    Please use "get_links" function in Oracle object to generate Links object.

    Attributes:
        links_dict (dictionary): Dictionary that store unprocessed network data.
        filtered_links (dictionary): Dictionary that store filtered network data.
        merged_score (pandas.dataframe): Network scores.
        cluster (list of str): List of cluster name.
        name (str): Name of clustering unit.
        palette (pandas.dataframe): DataFrame that store color information.


    """

    def __init__(self, name, links_dict={}):
        """
        Initialize Links object.

        Args:
            name (str): Name for linkLists.
            links_dict (dictionary): Python dictionary that store linkLists.
                Keys are cluster name, values are linkLists.
        """


        self.name = name
        self.links_dict = links_dict.copy()
        self.cluster = list(self.links_dict.keys())
        self.entropy = None

    def to_hdf5(self, file_path):
        """
        Save object as hdf5.

        Args:
            file_path (str): file path to save file. Filename needs to end with '.celloracle.links'
        """
        if file_path.endswith(".celloracle.links"):
            pass
        else:
            raise ValueError("Filename needs to end with '.celloracle.links'")

        compression_opts = 7
        dump_hdf5(obj=self, filename=file_path,
                  data_compression=compression_opts,  chunks=(2048, 2048),
                  noarray_compression=compression_opts, pickle_protocol=2)

    def _pipeline(self):
        """
        Perform several processing methods below.
        [self.merge_links(), self.filter_links(), self.get_score()]

        """
        self.filter_links()
        self.get_score()


    def filter_links(self, p=0.001, weight="coef_abs",
                     threshold_number=10000,
                     genelist_source=None,
                     genelist_target=None,
                     thread_number=None):
        """

        Filter network edges.
        In most cases, inferred GRN has non-significant random edges.
        We have to remove these edges before analyzing the network structure.
        You can do the filtering in any of the following ways.

        (1) Filter based on the p-value of the network edge.
            Please enter p-value for thresholding.
        (2) Filter based on network edge number.
            If you set the number, network edges will be filtered based on the order of a network score. The top n-th network edges with network weight will remain, and the other edges will be removed.
            The network data has several types of network weight, so you have to select which network weight do you want to use.
        (3) Filter based on an arbitrary gene list. You can set a gene list for source nodes or target nodes.

        Args:
            p (float): threshold for p-value of the network edge.
            weight (str): Please select network weight name for the filtering
            genelist_source (list of str): gene list to remain in regulatory gene nodes. Default is None.
            genelist_target (list of str): gene list to remain in target gene nodes. Default is None.

        """
        if thread_number is not None:
            threshold_number = thread_number

        self.filtered_links = {}
        self.threshold_number=threshold_number
        for i in self.cluster:
            self.filtered_links[i] = _thresholding(
                            linkList=self.links_dict[i],
                            p=p, weight=weight,
                            threshold_number=threshold_number,
                            genelist_source=genelist_source,
                            genelist_target=genelist_target)




    def get_network_score(self):

        """
        Get several network sores using igraph library.
        The following scores are calculated: ['degree_all', 'degree_centrality_all', 'degree_in',
       'degree_centrality_in', 'degree_out', 'degree_centrality_out',
       'betweenness_centrality', 'eigenvector_centrality']

        """
        if hasattr(self, "filtered_links"):
            if isinstance(self.filtered_links, dict):
                pass
            else:
                raise ValueError("Filtered network was not found. Prease run 'filter_links' first.")
        else:
            raise ValueError("Filtered network was not found. Prease run 'filter_links' first.")

        network_scores = []
        for key, val in self.filtered_links.items():
            df = _get_network_score(filtered_linklist_df=val)
            df["cluster"] = key
            network_scores.append(df)
        self.merged_score = pd.concat(network_scores, axis=0)

    def _get_network_score_if_not_exist(self):
        if hasattr(self, "merged_score"):
            if isinstance(self.merged_score, pd.core.frame.DataFrame):
                pass
            else:
                warnings.warn("\nNetwork score was not found. Celloracle will try to calculate network score first.",
                              UserWarning)
                self.get_network_score()
        else:
            warnings.warn("\nNetwork score was not found. Celloracle will try to calculate network score first.",
                          UserWarning)
            self.get_network_score()

    def get_score(self, test_mode=False, n_jobs=-1):
        """
        Get several network sores using R-igraph, linkcomm, and rnetcarto.
        This require R packages.
        """
        warnings.warn(
            "This is the deprecated function. Please use 'get_network_score' instead. This function will be removed in the future version.",
            DeprecationWarning
        )

        if _check_R_libraries_installation():
            pass
        else:
            raise ValueError("This function is deprecated. Please use 'get_network_score' instead. If you still use this function, you need to install R packages. Please make sure these R packages are installed: 'igraph', 'linkcomm', 'rnetcarto'.")


        li = list(self.filtered_links.keys()) # make list of cluster name

        # make dictionary. we make unique id for each cluster and use it for temporary file name.
        id_dict = {}
        for id_, i in enumerate(li):
            id_dict[i] = id_


        _get_network_score_by_Rscripts_inparallel(
            dict_links=self.filtered_links,
            id_dict=id_dict,
            output_folder="network_analysis",
            message=False,
            n_parallel=n_jobs)

        network_scores = {}
        for i in li:
            network_scores[i] = _load_network_analysis_results(f"./network_analysis/{id_dict[i]}")
        self.merged_score = _merge_df(network_scores)

        if not test_mode:
            os.system(f"rm -r ./network_analysis/")

        #print(f"the scores are saved in ./{self.name}/")


    def get_network_entropy(self, value="coef_abs"):
        """
        Calculate network entropy scores.

        Args:
            value (str): Default is "coef_abs".
        """
        li = []
        for i in self.cluster:
            mat = _link2mat(self.links_dict[i], value)
            df = _getNetworkEntropy(mat)
            df["cluster"] = i
            #df = df.dropna(axis=0)
            li.append(df)
        self.entropy = pd.concat(li, axis=0)


    #####################################################
    #### Methods for network structure visualization ####
    #####################################################
    def plot_degree_distributions(self, plot_model=False, save=None):
        """
        Plot the network degree distributions (the number of edge per gene).
        The network degree will be visualized in both linear scale and log scale.

        Args:
            links (Links): See network_analysis.Links class for detail.
            plot_model (bool): Whether to plot linear approximation line.
            save (str): Folder path to save plots. If the folder does not exist in the path, the function creates the folder.
               Plots will not be saved if [save=None]. Default is None.

        """
        if hasattr(self, "filtered_links"):
            if isinstance(self.filtered_links, dict):
                pass
            else:
                raise ValueError("Filtered network was not found. Prease run 'filter_links' first.")
        else:
            raise ValueError("Filtered network was not found. Prease run 'filter_links' first.")

        plot_degree_distributions(links=self, plot_model=plot_model, save=save)


    def plot_score_discributions(self, values=None, method="boxplot", save=None):
        """
        Plot the distribution of network scores.
        An individual data point is a network edge (gene).

        Args:
            links (Links): See Links class for details.
            values (list of str): The list of score to visualize. If it is None, all of the network score will be used.
            method (str): Plotting method. Select either "boxplot" or "barplot".
            save (str): Folder path to save plots. If the folder does not exist in the path, the function creates the folder.
               Plots will not be saved if [save=None]. Default is None.
        """
        self._get_network_score_if_not_exist()
        plot_score_discributions(links=self, values=values, method=method, save=save)

    def plot_network_entropy_distributions(self, update_network_entropy=False, save=None):
        """
        Plot the distribution for network entropy.
        See the CellOracle paper for more detail.

        Args:
            links (Links object): See network_analysis.Links class for detail.
            values (list of str): The list of score to visualize. If it is None, all network score (listed above) will be used.
            update_network_entropy (bool): Whether to recalculate network entropy.
            save (str): Folder path to save plots. If the folder does not exist in the path, the function creates the folder.
               Plots will not be saved if [save=None]. Default is None.
        """
        plot_network_entropy_distributions(links=self, update_network_entropy=update_network_entropy, save=save)


    ###################################
    #### Methods for gene analysis ####
    ###################################
    def plot_scores_as_rank(self, cluster, n_gene=50, save=None):
        """
        Pick up top n-th genes wich high-network scores and make plots.

        Args:
            links (Links): See network_analysis.Links class for detail.
            cluster (str): Cluster name to analyze.
            n_gene (int): Number of genes to plot. Default is 50.
            save (str): Folder path to save plots. If the folder does not exist in the path, the function creates the folder.
               Plots will not be saved if [save=None]. Default is None.
        """
        self._get_network_score_if_not_exist()

        plot_scores_as_rank(links=self, cluster=cluster, n_gene=n_gene, save=save)

    def plot_score_comparison_2D(self, value, cluster1, cluster2, percentile=99, annot_shifts=None, save=None, plt_show=True, interactive=False):
        """
        Make a scatter plot that compares specific network scores in two groups.

        Args:
            links (Links): See network_analysis.Links class for detail.
            value (srt): The network score type.
            cluster1 (str): Cluster name. Network scores in cluster1 will be visualized in the x-axis.
            cluster2 (str): Cluster name. Network scores in cluster2 will be visualized in the y-axis.
            percentile (float): Genes with a network score above the percentile will be shown with annotation. Default is 99.
            annot_shifts ((float, float)): Annotation visualization setting.
            save (str): Folder path to save plots. If the folder does not exist in the path, the function creates the folder.
               Plots will not be saved if [save=None]. Default is None.
        """
        self._get_network_score_if_not_exist()

        if interactive:
            return plot_score_comparison_2D_with_plotly(links=self, value=value, cluster1=cluster1, cluster2=cluster2, fillna_with_zero=True)
        else:
            plot_score_comparison_2D(links=self, value=value, cluster1=cluster1, cluster2=cluster2,
                                     percentile=percentile, annot_shifts=annot_shifts, save=save, plt_show=plt_show)

    def plot_score_per_cluster(self, goi, save=None, plt_show=True):
        """
        Plot network score for a gene.
        This function visualizes the network score for a specific gene between clusters to get an insight into the dynamics of the gene.

        Args:
            links (Links): See network_analysis.Links class for detail.
            goi (srt): Gene name.
            save (str): Folder path to save plots. If the folder does not exist in the path, the function creates the folder.
               Plots will not be saved if [save=None]. Default is None.
        """
        self._get_network_score_if_not_exist()

        plot_score_per_cluster(links=self, goi=goi, save=save, plt_show=plt_show)

    def plot_cartography_scatter_per_cluster(self, gois=None, clusters=None,
                                             scatter=True, kde=False,
                                             auto_gene_annot=False, percentile=98,
                                             args_dot={"n_levels": 105}, args_line={"c":"gray"},
                                             args_annot={}, save=None):

        """
        Make a gene network cartography plot.
        Please read the original paper describing gene network cartography for more information.
        https://www.nature.com/articles/nature03288

        Args:
            links (Links): See network_analysis.Links class for detail.
            gois (list of srt): List of gene name to highlight.
            clusters (list of str): List of cluster name to analyze. If None, all clusters in Links object will be analyzed.
            scatter (bool): Whether to make a scatter plot.
            auto_gene_annot (bool): Whether to pick up genes to make an annotation.
            percentile (float): Genes with a network score above the percentile will be shown with annotation. Default is 98.
            args_dot (dictionary): Arguments for scatter plot.
            args_line (dictionary): Arguments for lines in cartography plot.
            args_annot (dictionary): Arguments for annotation in plots.
            save (str): Folder path to save plots. If the folder does not exist in the path, the function creates the folder.
               Plots will not be saved if [save=None]. Default is None.

        """
        warnings.warn(
            "This is the deprecated function. This function will be removed in the future version.",
            DeprecationWarning
        )

        if "role" not in self.merged_score.columns:
            print("Cartography is not calculated yet. Please run 'get_score' first. This function require several R libraries.")
            return None

        plot_cartography_scatter_per_cluster(links=self, gois=gois, clusters=clusters,
                                             scatter=scatter, kde=kde,
                                             auto_gene_annot=auto_gene_annot, percentile=percentile,
                                             args_dot=args_dot, args_line=args_line,
                                             args_annot=args_annot, save=save)

    def plot_cartography_term(self, goi, save=None, plt_show=True):


        """
        Plot the gene network cartography term like a heatmap.
        Please read the original paper of gene network cartography for the principle of gene network cartography.
        https://www.nature.com/articles/nature03288

        Args:
            links (Links): See network_analysis.Links class for detail.
            gois (list of srt): List of gene name to highlight.
            save (str): Folder path to save plots. If the folder does not exist in the path, the function creates the folder.
               Plots will not be saved if [save=None]. Default is None.
        """
        warnings.warn(
            "This is the deprecated function. This function will be removed in the future version.",
            DeprecationWarning
        )

        if "role" not in self.merged_score.columns:
            print("Cartography is not calculated yet. Please run 'get_score' first. This function require several R libraries.")
            return None



        plot_cartography_term(links=self, goi=goi, save=save, plt_show=plt_show)


def _link2mat(link, value="coef_abs", fillna=0):
    mat = pd.pivot(data=link,
                   values=[value],
                   index="target", columns="source")
    mat = mat.fillna(fillna)

    return mat


def _getNetworkEntropy(linkMat):
    tmp = linkMat.copy()
    k = (tmp != 0).sum(axis=1)
    tmp = linkMat.copy()
    ent = []
    ent_norm = []
    genes = []
    for i in tmp.index:
        if k[i] > 1:
            en = stats.entropy(tmp.loc[i])
            ent.append(en)
            ent_norm.append(en/np.log(k[i]))
            genes.append(i)

    df = pd.DataFrame({"entropy": ent, "entropy_norm": ent_norm},
                     index=genes).dropna(axis=0)
    return df




def _thresholding(linkList, p=None, weight="coef_abs",
                 threshold_number=10000, genelist_source=None,
                genelist_target=None):
    li = linkList.copy()
    if not genelist_source is None:
        li = li[li.source.isin(genelist_source)]
    if not genelist_target is None:
        li = li[li.target.isin(genelist_target)]

    if not p is None:
        li = li[li["p"] <= p]
    li = li.sort_values(weight, ascending=False)

    if not threshold_number is None:
        li = li[:threshold_number]

    #li = li[["source", "target", weight]]


    return li

def _load_network_analysis_results(folder):

    files = os.listdir(folder)
    network_score = pd.read_csv(os.path.join(folder, "base_natwork_analysis.csv"), index_col=0)
    #overlapping_cluster = pd.read_csv(os.path.join(folder, "overlapping_cluster.csv"), index_col=0)
    #if "overlapping_cluster_GO.csv" in files:
     #   overlapping_cluster_GO = pd.read_csv(os.path.join(folder, "overlapping_cluster_GO.csv"), index_col=0)
    #    return network_score, overlapping_cluster, overlapping_cluster_GO
    #else:
    return network_score#, overlapping_cluster


def _merge_df(link_dict):

    merged = []
    clusters = link_dict.keys()
    for i in clusters:
        tmp = link_dict[i].copy()
        tmp["cluster"] = i
        merged.append(tmp)
    merged = pd.concat(merged, axis=0)

    return merged

def _get_network_score(filtered_linklist_df):
    """
    This is the function to get network score for each node.
    This function is made to reproduce previous functions made with R-igraph.
    """
    # Make igraph object
    g = Graph.DataFrame(filtered_linklist_df[["source", "target"]],
                        directed=True,
                        use_vids=False)
    g.es["weight"] = filtered_linklist_df["coef_abs"].values.copy()

    # Make placeholder
    df = g.get_vertex_dataframe()

    # Calculate scores
    for i in ["all", "in", "out"]:
        df[f"degree_{i}"] = g.degree(mode=i)
        df[f"degree_centrality_{i}"] = df[f"degree_{i}"] / (df.shape[0]-1)
    df["betweenness_centrality"] = g.betweenness(directed=True, weights="weight")
    df["eigenvector_centrality"] = g.eigenvector_centrality(directed=False, weights="weight")

    # Use gene name for index
    df = df.set_index("name")
    df.index.name = None

    return df



# -*- coding: utf-8 -*-



import os
import sys
import warnings

import pandas as pd
import numpy as np

from sklearn.linear_model import Ridge
from sklearn.neighbors import KNeighborsRegressor, KNeighborsClassifier
from sklearn.preprocessing import PolynomialFeatures


def scatter_value_to_grid_value(embedding, grid, value, method="knn", n_knn=30, n_poly=3):

    """
    Transfer value on 2D scatter into 2d Grid space.

    Args:

        embedding (np.array): shape = (n_cell, 2)
        grid (np.array): shape = (n_grid**2, 2)
        value (np.array): shape = (n_cell, )

    Returns:
        np.array : shape = (n_grid**2, 1)

    """

    x, y = embedding[:, 0], embedding[:, 1]
    x_new, y_new = grid[:, 0], grid[:, 1]

    if method == "poly":
        value_on_grid =  _polynomial_regression_old_ver(x, y, x_new, y_new, value, n_degree=n_poly)
    if method == "polynomial":
        value_on_grid =  _polynomial_regression_sklearn(x, y, x_new, y_new, value, n_degree=n_poly)
    elif method == "knn":
        value_on_grid = _knn_regression(x, y, x_new, y_new, value, n_knn=n_knn)
    elif method == "knn_class":
        value_on_grid = _knn_classification(x, y, x_new, y_new, value, n_knn=n_knn)


    return value_on_grid


def _polynomial_regression_sklearn(x, y, x_new, y_new, value, n_degree=3):

    # Make polynomial features
    data = np.stack([x, y], axis=1)
    data_new = np.stack([x_new, y_new], axis=1)

    pol = PolynomialFeatures(degree=n_degree, include_bias=False)
    data = pol.fit_transform(data)
    data_new = pol.transform(data_new)

    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        model = Ridge(random_state=123)
        model.fit(data, value)

    return model.predict(data_new)

def _polynomial_regression_old_ver(x, y, x_new, y_new, value, n_degree=3):

    def __conv(x, y, n_degree=3): # Make polynomial data for polynomial ridge regression
        dic = {}
        for d in range(1, n_degree + 1):
            for i, name in zip([x, y], ["x", "y"]):
                dic[name + str(d)] = i**d
        return pd.DataFrame(dic)

    data = __conv(x=x, y=y, n_degree=n_degree)

    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        model = Ridge()
        model.fit(data, value)

    data_new = __conv(x=x_new, y=y_new, n_degree=n_degree)

    return model.predict(data_new)

def _knn_regression(x, y, x_new, y_new, value, n_knn=30):

    data = np.stack([x, y], axis=1)

    model = KNeighborsRegressor(n_neighbors=n_knn)
    model.fit(data, value)

    data_new = np.stack([x_new, y_new], axis=1)

    return model.predict(data_new)

def _knn_classification(x, y, x_new, y_new, value, n_knn=30):

    data = np.stack([x, y], axis=1)

    model = KNeighborsClassifier(n_neighbors=n_knn)
    model.fit(data, value)

    data_new = np.stack([x_new, y_new], axis=1)

    return model.predict(data_new)



# -*- coding: utf-8 -*-

import sys
import re
import warnings
import logging

from . import utility, network, network_analysis, go_analysis, data, data_conversion, oracle_utility
from .trajectory.oracle_core import Oracle
from .network import Net
from .network_analysis import Links
from .utility.load_hdf5 import load_hdf5
from .utility import check_python_requirements
from .network_analysis import test_R_libraries_installation
from .version import __version__
#from . import motif_analysis

logger = logging.getLogger(__name__)
logger.addHandler(logging.StreamHandler())
logger.setLevel(logging.INFO)


# Make sure that DeprecationWarning within this package always gets printed
warnings.filterwarnings('always', category=DeprecationWarning,
                        module=r'^{0}\.'.format(re.escape(__name__)))

__copyright__    = 'Copyright (C) 2020 Kenji Kamimoto'
__license__      = 'Apache License Version 2.0'
__author__       = 'Kenji Kamimoto'
__author_email__ = 'kamimoto@wustl.edu'
__url__          = 'https://github.com/morris-lab/CellOracle'
#__version__ = "0.10.2"
__all__ = ["utility", "motif_analysis", "network", "network_analysis",
           "go_analysis", "data", "data_conversion",
           "Oracle",
           "Links",
           "Net",
           "load_hdf5",
           "test_R_libraries_installation",
           "check_python_requirements"
           ]

check_python_requirements(return_detail=False, print_warning=True)



# -*- coding: utf-8 -*-



import io
import logging
import os
import pickle
import subprocess
import sys
from copy import deepcopy

import pandas as pd
import numpy as np
from tqdm.auto import tqdm
import matplotlib.pyplot as plt
import seaborn as sns
#import h5py

from scipy.stats import wilcoxon

from ..trajectory.oracle_core import Oracle
from .scatter_to_grid import scatter_value_to_grid_value
from .utility import Data_strage

from ..visualizations.config import CONFIG
from ..visualizations.development_module_visualization import (\
                            plot_cluster_whole,
                            plot_cluster_cells_use,
                            plot_background,
                            plot_pseudotime,
                            plot_pseudotime_on_grid,
                            plot_selected_pseudotime_on_grid,
                            plot_reference_flow_on_grid,
                            plot_simulation_flow_on_grid,
                            plot_simulation_flow_random_on_grid,
                            plot_inner_product_on_grid,
                            plot_inner_product_random_on_grid,
                            plot_inner_product_on_pseudotime,
                            plot_inner_product_as_box,
                            plot_quiver,
                            plot_quiver_random,
                            visualize_development_module_layout_0,
                            visualize_development_module_layout_1,
                            visualize_development_module_layout_2)

class Oracle_development_module(Data_strage):

    def __init__(self, oracle_object=None, gradient_object=None, name=None):
        super().__init__()
        self._exemptions_when_del_attrs = []

        self.name = name
        self.cell_idx_use = None

        if gradient_object is not None:
            self.load_differentiation_reference_data(gradient_object=gradient_object)

        if oracle_object is not None:
            self.load_perturb_simulation_data(oracle_object=oracle_object)

    def copy(self):
        return deepcopy(self)

    def set_hdf_path(self, path):
        key = "" # placeholder
        self._set_hdf_path(path=path, key=key, create_if_not_exist=True)

    def dump_hdf5(self, gene, misc):

        if hasattr(self, "_path"):
            pass
        else:
            raise ValueError("Run set_hdf_path to set file path first.")

        self._dump_hdf5(place=f"{gene}/{misc}")


    def load_hdf5(self, gene, misc, specify_attributes=None):

        self._load_hdf5(place=f"{gene}/{misc}", specify_attributes=specify_attributes)

    def del_attrs(self, exemptions=[]):
        # Delete all attributes
        self._del_attrs(exemptions=exemptions)

    def get_hdf5_info(self):

        dic = {}
        # Gene list
        dic["gene_list"] = np.unique([i.split("/")[0] for i in self._names])
        # Misc list
        dic["misc_list"] = np.unique([i.split("/")[1] for i in self._names if len(i.split("/")) >= 2])
        #
        unique_keys = np.unique(["/".join(i.split("/")[:2]) for i in self._names if len(i.split("/")) >= 2])
        dic["gene_misc_lists"] = list(map(lambda x: x.split("/"), unique_keys))

        # Get dictionary; key = misc, value = list of gene
        misc_gene_dictionary = {}
        for i in dic["misc_list"]:
            misc_gene_dictionary[i] = [gene for gene, misc in dic["gene_misc_lists"] if misc == i]
        dic["misc_gene_dictionary"] = misc_gene_dictionary

        return dic


    def load_differentiation_reference_data(self, gradient_object):

        self.n_grid = gradient_object.n_grid
        self.min_mass = gradient_object.min_mass
        self.smooth = gradient_object.smooth
        self.n_neighbors = gradient_object.n_neighbors

        self.pseudotime = gradient_object.pseudotime.copy() # shape = (n_cell, )
        self.pseudotime_on_grid = gradient_object.pseudotime_on_grid.copy() # shape = (n_grid*n_grid, )
        self.mass_filter_reference = gradient_object.mass_filter.copy() # shape = (n_grid*n_grid, )
        self.mass_filter_whole_reference = gradient_object.mass_filter_whole.copy() # shape = (n_grid*n_grid, )
        self.gridpoints_coordinates = gradient_object.gridpoints_coordinates.copy() # shape = (n_grid*n_grid, 2)

        self.ref_flow = gradient_object.ref_flow.copy() # shape = (n_grid*n_grid, 2)

        if gradient_object.cell_idx_use is not None:
            self.cell_idx_use = gradient_object.cell_idx_use.copy()

    def load_perturb_simulation_data(self, oracle_object, cell_idx_use=None, name=None, min_mass=None, n_neighbors=None, smooth=None):

        if not hasattr(self, "pseudotime"):
            raise ValueError("Please load differentiation reference data first.")

        # 0. Store data
        for i in ["delta_embedding", "delta_embedding_random", "embedding", "colorandum"]:
            setattr(self, i, getattr(oracle_object, i))

        if cell_idx_use is not None:
            self.cell_idx_use = np.array(cell_idx_use)

        if name is not None:
            self.name = name

        # 1. Subset oracle data if cell_idx_use is specified
        if cell_idx_use is None:
            oracle_object_ = oracle_object
        else:
            oracle_object_ = subset_oracle_for_development_analysiis(oracle_object=oracle_object,
                                                          cell_idx_use=cell_idx_use)

        # 2. Grid calculation
        if n_neighbors is None:
            n_neighbors = self.n_neighbors
        if smooth is None:
            smooth = self.smooth
        x_min, y_min = self.embedding.min(axis=0)
        x_max, y_max = self.embedding.max(axis=0)
        xylim = ((x_min, x_max), (y_min, y_max))

        oracle_object_.calculate_grid_arrows(smooth=smooth, steps=(self.n_grid, self.n_grid), n_neighbors=n_neighbors, xylim=xylim)


        # 3. Store result of grid calculation
        for i in ["flow", "flow_rndm"]:
            setattr(self, i, getattr(oracle_object_, i))

        if not (oracle_object_.flow_grid == self.gridpoints_coordinates).all():
            raise ValueError("Grid point cordinates are not mached.")

        # 4. mass_filter calculation
        if min_mass is None:
            min_mass = self.min_mass
        self.mass_filter_simulation = (oracle_object_.total_p_mass < min_mass)


    def calculate_inner_product(self):

        # Calculate inner product between the pseudotime-gradient and the perturb-gradient
        self.inner_product = np.array([np.dot(i, j) for i, j in zip(self.flow, self.ref_flow)])
        self.inner_product_random = np.array([np.dot(i, j) for i, j in zip(self.flow_rndm, self.ref_flow)])


    def calculate_digitized_ip(self, n_bins=10):

        inner_product_df = pd.DataFrame({"score": self.inner_product[~self.mass_filter_simulation],
                                         "score_randomized": self.inner_product_random[~self.mass_filter_simulation],
                                         "pseudotime": self.pseudotime_on_grid[~self.mass_filter_simulation]})


        bins = _get_bins(inner_product_df.pseudotime, n_bins)
        inner_product_df["pseudotime_id"] = np.digitize(inner_product_df.pseudotime, bins)

        self.inner_product_df = inner_product_df

    def get_negative_PS_p_value(self, pseudotime=None, return_ps_sum=False, plot=False):

        df = self.inner_product_df.copy()

        if pseudotime is not None:
            pseudotime = [i for i in pseudotime if i in list("0123456789")]
            df = df[df.pseudotime_id.isin(pseudotime)]

        x = df["score"]
        y = df["score_randomized"]

        # Clipping positive value to focus on negative IP
        x = np.clip(x, -np.inf, 0)
        y = np.clip(y, -np.inf, 0)

        if plot:
            sns.distplot(x)
            sns.distplot(y)

        # Paired non-parametric test with Wilcoxon's runk sum test
        s, p = wilcoxon(x, y, alternative="less")

        if return_ps_sum:
            return p, -x.sum(), -y.sum()

        return p

    def get_positive_PS_p_value(self, pseudotime=None, return_ps_sum=False, plot=False):
        df = self.inner_product_df.copy()

        if pseudotime is not None:
            pseudotime = [i for i in pseudotime if i in list("0123456789")]
            df = df[df.pseudotime_id.isin(pseudotime)]

        x = df["score"]
        y = df["score_randomized"]

        # Clipping negative value to focus on positive IP
        x = np.clip(x, 0, np.inf)
        y = np.clip(y, 0, np.inf)

        if plot:
            sns.distplot(x)
            sns.distplot(y)

        # Paired non-parametric test with Wilcoxon's runk sum test
        s, p = wilcoxon(x, y, alternative="greater")

        if return_ps_sum:
            return p, x.sum(), y.sum()

        return p

    def get_sum_of_negative_ips(self):
        return get_sum_of_negative_ips(inner_product_df=self.inner_product_df)

    def get_sum_of_positive_ips(self):
        return get_sum_of_positive_ips(inner_product_df=self.inner_product_df)

    # Visualization
    def plot_cluster_whole(self, ax=None, s=CONFIG["s_scatter"], args=CONFIG["default_args"]):
        plot_cluster_whole(self=self, ax=ax, s=s, args=args)

    def plot_cluster_cells_use(self, ax=None, s=CONFIG["s_scatter"], color=None, show_background=True, args=CONFIG["default_args"]):
        plot_cluster_cells_use(self=self, ax=ax, s=s, color=color, show_background=show_background, args=args)

    def plot_background(self, ax=None, s=CONFIG["s_scatter"], args=CONFIG["default_args"]):
        plot_background(self=self, ax=ax, s=s, args=args)

    def plot_pseudotime(self, ax=None, s=CONFIG["s_scatter"], show_background=True, cmap="rainbow", args=CONFIG["default_args"]):
        plot_pseudotime(self=self, ax=ax, s=s, show_background=show_background, cmap=cmap, args=args)

    def plot_background_on_grid(self, ax=None, s=CONFIG["s_grid"], args={}):
        plot_background_on_grid(self=self, ax=ax, s=s, args=args)

    def plot_pseudotime_on_grid(self, ax=None, s=CONFIG["s_grid"], show_background=True, cmap="rainbow", args={}):
        plot_pseudotime_on_grid(self=self, ax=ax, s=s, show_background=show_background, cmap=cmap, args=args)

    def plot_selected_pseudotime_on_grid(self, ax=None, pseudotime_selected=[], s=CONFIG["s_grid"], show_background=True, args={}):
        plot_selected_pseudotime_on_grid(self=self, ax=ax, pseudotime_selected=pseudotime_selected, s=s, show_background=show_background, args=args)

    def plot_reference_flow_on_grid(self, ax=None, scale=CONFIG["scale_dev"], show_background=True, s=CONFIG["s_scatter"], args=CONFIG["default_args_quiver"]):
        plot_reference_flow_on_grid(self=self, ax=ax, scale=scale, show_background=show_background, s=s, args=args)

    def plot_simulation_flow_on_grid(self, ax=None, scale=CONFIG["scale_simulation"], show_background=True, s=CONFIG["s_scatter"], args=CONFIG["default_args_quiver"]):
        plot_simulation_flow_on_grid(self, ax=ax, scale=scale, show_background=show_background, s=s, args=args)

    def plot_simulation_flow_random_on_grid(self, ax=None, scale=CONFIG["scale_simulation"], show_background=True, s=CONFIG["s_scatter"], args=CONFIG["default_args_quiver"]):
        plot_simulation_flow_random_on_grid(self, ax=ax, scale=scale, show_background=show_background, s=s, args=args)

    def plot_inner_product_on_grid(self, ax=None, vm=1, s=CONFIG["s_grid"], show_background=True, vmin=None, vmax=None, cmap=None, args={}):
        plot_inner_product_on_grid(self=self, ax=ax, vm=vm, s=s, show_background=show_background, vmin=vmin, vmax=vmax, cmap=cmap, args=args)

    def plot_inner_product_random_on_grid(self, ax=None, vm=1, s=CONFIG["s_grid"], show_background=True, vmin=None, vmax=None, cmap=None, args={}):
        plot_inner_product_random_on_grid(self=self, ax=ax, vm=vm, s=s, show_background=show_background, vmin=vmin, vmax=vmax, cmap=cmap, args=args)

    def plot_inner_product_on_pseudotime(self, ax=None, vm=1, s=CONFIG["s_grid"], vmin=None, vmax=None, cmap=None, args={}):
        plot_inner_product_on_pseudotime(self=self, ax=ax, vm=vm, s=s, vmin=vmin, vmax=vmax, cmap=cmap, args=args)

    def plot_inner_product_as_box(self, ax=None, vm=1, vmin=None, vmax=None, args={}):
        plot_inner_product_as_box(self=self, ax=ax, vm=vm, vmin=vmin, vmax=vmax, args=args)

    def plot_quiver(self, ax=None, scale=CONFIG["scale_simulation"], color=None, s=CONFIG["s_scatter"], show_background=True, args=CONFIG["default_args"]):
        plot_quiver(self=self, ax=ax, scale=scale, color=color, s=s, show_background=show_background, args=args)

    def plot_quiver_random(self, ax=None, scale=CONFIG["scale_simulation"], color=None, s=CONFIG["s_scatter"], show_background=True, args=CONFIG["default_args"]):
        plot_quiver_random(self=self, ax=ax, scale=scale, color=color, s=s, show_background=show_background, args=args)

    def visualize_development_module_layout_2(self, scale_for_pseudotime=CONFIG["scale_dev"],
        scale_for_simulation=CONFIG["scale_simulation"], s=CONFIG["s_scatter"], s_grid=CONFIG["s_grid"], vm=1, show_background=True, return_fig=False):

        fig = visualize_development_module_layout_2(self=self, scale_for_pseudotime=scale_for_pseudotime,
            scale_for_simulation=scale_for_simulation, s=s, s_grid=s_grid, vm=vm, show_background=show_background)

        if return_fig:
            return fig

    def visualize_development_module_layout_1(self, scale_for_pseudotime=CONFIG["scale_dev"],
        scale_for_simulation=CONFIG["scale_simulation"], s=CONFIG["s_scatter"], s_grid=CONFIG["s_grid"], vm=1, show_background=True, return_fig=False):

        fig = visualize_development_module_layout_1(self=self, scale_for_pseudotime=scale_for_pseudotime,
            scale_for_simulation=scale_for_simulation, s=s, s_grid=s_grid, vm=vm, show_background=show_background)

        if return_fig:
            return fig

    def visualize_development_module_layout_0(self, scale_for_pseudotime=CONFIG["scale_dev"],
        scale_for_simulation=CONFIG["scale_simulation"], s=CONFIG["s_scatter"], s_grid=CONFIG["s_grid"], vm=1, show_background=True, return_fig=False):

        fig = visualize_development_module_layout_0(self=self, scale_for_pseudotime=scale_for_pseudotime,
            scale_for_simulation=scale_for_simulation, s=s, s_grid=s_grid, vm=vm, show_background=show_background)

        if return_fig:
            return fig


def subset_oracle_for_development_analysiis(oracle_object, cell_idx_use):
    """
    Make a subset of oracle object by specifying of cluster.
    This function pick up some of attributes that needed for development analysis rather than whole attributes.

    """

    # Create new oracle object and transfer data
    oracle_ = Oracle()
    for i in ["embedding", "delta_embedding", "delta_embedding_random", "corrcoef_random", "adata"]:
        setattr(oracle_, i, getattr(oracle_object, i))

    for i in ["embedding", "delta_embedding", "delta_embedding_random", "corrcoef_random"]:
        setattr(oracle_, i, getattr(oracle_, i)[cell_idx_use])

    #cells_of_interest = oracle_object.adata.obs.index.values[cell_idx_use]
    #oracle_.adata = oracle_.adata[cells_of_interest, :]

    return oracle_

"""def _get_bins(array, n_bins):
    min_ = array.min()
    max_ = array.max()
    width = (max_ - min_)/(n_bins-1)
    return np.arange(min_, max_ + width, width)"""

def _get_bins(array, n_bins):
    min_ = array.min()
    max_ = array.max()
    width = (max_ - min_)/(n_bins)
    return np.arange(min_, max_ + width, width)[1:-1]


from scipy.stats import wilcoxon



def get_sum_of_positive_ips(inner_product_df):

    df = inner_product_df[["pseudotime_id", "score"]][inner_product_df.score > 0]
    df = df.groupby(by="pseudotime_id").sum()
    df = df.reset_index(drop=False)

    return df

def get_sum_of_negative_ips(inner_product_df):

    df = inner_product_df[["pseudotime_id", "score"]][inner_product_df.score < 0]
    df = df.groupby(by="pseudotime_id").sum()
    df = df.reset_index(drop=False)

    return df



# -*- coding: utf-8 -*-


import os, sys

import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
import seaborn as sns

from .config import CONFIG

from .development_module_visualization import (plot_cluster_whole,
                            plot_quiver, plot_quiver_random,
                            plot_simulation_flow_on_grid,
                            plot_simulation_flow_random_on_grid)

class Oracle_visualization:
    """docstring for ."""

    def __init__(self):
        pass

    def plot_cluster_whole(self, ax=None, s=CONFIG["s_scatter"], args=CONFIG["default_args"]):
        plot_cluster_whole(self=self, ax=ax, s=s, args=args)

    def plot_quiver(self, ax=None, scale=CONFIG["scale_simulation"], color=None, s=CONFIG["s_scatter"], show_background=True, args=CONFIG["default_args"]):
        plot_quiver(self=self, ax=ax, scale=scale, color=color, s=s, show_background=show_background, args=args)

    def plot_quiver_random(self, ax=None, scale=CONFIG["scale_simulation"], color=None, s=CONFIG["s_scatter"], show_background=True, args=CONFIG["default_args"]):
        plot_quiver_random(self=self, ax=ax, scale=scale, color=color, s=s, show_background=show_background, args=args)

    def plot_simulation_flow_on_grid(self, ax=None, scale=CONFIG["scale_simulation"], show_background=True, s=CONFIG["s_scatter"], args=CONFIG["default_args_quiver"]):
        plot_simulation_flow_on_grid(self=self, ax=ax, scale=scale, show_background=show_background, s=s, args=args)

    def plot_simulation_flow_random_on_grid(self, ax=None, scale=CONFIG["scale_simulation"], show_background=True, s=CONFIG["s_scatter"], args=CONFIG["default_args_quiver"]):
        plot_simulation_flow_random_on_grid(self=self, ax=ax, scale=scale, show_background=show_background, s=s, args=args)



# -*- coding: utf-8 -*-
"""
The :mod:`.data` module implements data download and loading.

"""

from .load_data import (load_TFinfo_df_mm9_mouse_atac_atlas,
                        load_mouse_scATAC_atlas_base_GRN,
                        load_Paul2015_data,
                        load_tutorial_links_object,
                        load_tutorial_oracle_object)

from .load_promoter_base_GRN import (load_drosophila_promoter_base_GRN,
                                     load_rat_promoter_base_GRN,
                                     load_mouse_promoter_base_GRN,
                                     load_human_promoter_base_GRN,
                                     load_chicken_promoter_base_GRN,
                                     load_zebrafish_promoter_base_GRN,
                                     load_xenopus_tropicalis_promoter_base_GRN,
                                     load_xenopus_laevis_promoter_base_GRN,
                                     load_arabidopsis_promoter_base_GRN,
                                     load_Scerevisiae_promoter_base_GRN,
                                     load_Celegans_promoter_base_GRN,
                                     load_Pig_promoter_base_GRN)

__all__ = ["load_TFinfo_df_mm9_mouse_atac_atlas",
           "load_mouse_scATAC_atlas_base_GRN"
           "load_Paul2015_data",
           "load_tutorial_links_object",
           "load_tutorial_oracle_object",
           "load_drosophila_promoter_base_GRN",
           "load_rat_promoter_base_GRN",
           "load_mouse_promoter_base_GRN",
           "load_human_promoter_base_GRN",
           "load_chicken_promoter_base_GRN",
           "load_zebrafish_promoter_base_GRN",
           "load_xenopus_tropicalis_promoter_base_GRN",
           "load_xenopus_laevis_promoter_base_GRN"
           "load_arabidopsis_promoter_base_GRN",
           "load_Scerevisiae_promoter_base_GRN",
           "load_Celegans_promoter_base_GRN",
           "load_Pig_promoter_base_GRN"]



from copy import deepcopy
import warnings
import logging
import numpy as np

import pandas as pd
import os
from numba import jit
import scanpy as sc

from ..utility import intersect

import warnings

#########################
### utility functions ###
#########################


#################################################
### functions for data handling with anndata  ###
#################################################
def _check_color_information_and_create_if_not_found(adata, cluster_column_name, embedding_name):
    if f"{cluster_column_name}_colors" in adata.uns.keys():
        pass
    else:
        message = f"Color information for the {cluster_column_name} is not found in the anndata. CellOracle is plotting the clustering data, {cluster_column_name}, to create color data."
        warnings.warn(message, UserWarning)
        sc.pl.embedding(adata, basis=embedding_name, color=cluster_column_name)

def _linklist2dict(linklist):
    dic = {}
    tmp = linklist.set_index("target")
    for i in tmp.index.unique():
        dic[i] = tmp.loc[i][["source"]].values.flatten()
    return dic

def _adata_to_matrix(adata, layer_name, transpose=True):
    """
    Extract an numpy array from adata and returns as numpy matrix.

    Args:
        adata (anndata): anndata

        layer_name (str): name of layer in anndata

        trabspose (bool) : if True, it returns transposed array.

    Returns:
        2d numpy array: numpy array
    """
    if isinstance(adata.layers[layer_name], np.ndarray):
        matrix = adata.layers[layer_name].copy()
    else:
        matrix = adata.layers[layer_name].todense().A.copy()

    if transpose:
        matrix = matrix.transpose()

    return matrix.copy(order="C")


def _adata_to_df(adata, layer_name, transpose=False):
    """
    Extract an numpy array from adata and returns as pandas DataFrane with cell names and gene names.

    Args:
        adata (anndata): anndata

        layer_name (str): name of layer in anndata

        trabspose (bool) : if True, it returns transposed array.

    Returns:
        pandas.DataFrame: data frame (cells x genes (if transpose == False))
    """
    array = _adata_to_matrix(adata, layer_name, transpose=False)
    df = pd.DataFrame(array, columns=adata.var.index.values, index=adata.obs.index.values)

    if transpose:
        df = df.transpose()
    return df

def _adata_to_color_dict(adata, cluster_use):
    """
    Extract color information from adata and returns as dictionary.

    Args:
        adata (anndata): anndata

        cluster_use (str): column name in anndata.obs

    Returns:
        dictionary: python dictionary, key is cluster name, value is clor name
    """
    color_dict = {}
    for i,j in enumerate(adata.obs[cluster_use].cat.categories):
        color_dict[j] = adata.uns[f"{cluster_use}_colors"][i]
    return color_dict



def _get_clustercolor_from_anndata(adata, cluster_name, return_as):
    """
    Extract clor information from adata and returns as palette (pandas data frame) or dictionary.

    Args:
        adata (anndata): anndata

        cluster_name (str): cluster name in anndata.obs

        return_as (str) : "palette" or "dict"

    Returns:
        2d numpy array: numpy array
    """
        # return_as: "palette" or "dict"
    def float2rgb8bit(x):
        x = (x*255).astype("int")
        x = tuple(x)

        return x

    def rgb2hex(rgb):
        return '#%02x%02x%02x' % rgb

    def float2hex(x):
        x = float2rgb8bit(x)
        x = rgb2hex(x)
        return x

    def hex2rgb(c):
        return (int(c[1:3],16),int(c[3:5],16),int(c[5:7],16), 255)

    pal = get_palette(adata, cluster_name)
    if return_as=="palette":
        return pal
    elif return_as=="dict":
        col_dict = {}
        for i in pal.index:
            col_dict[i] = np.array(hex2rgb(pal.loc[i, "palette"]))/255
        return col_dict
    else:
        raise ValueErroe("return_as")
    return 0

def get_palette(adata, cname):
    c = [i.upper() for i in adata.uns[f"{cname}_colors"]]
    #c = sns.cubehelix_palette(24)
    """
    col = adata.obs[cname].unique()
    col = list(col)
    col.sort()
    """
    try:
        col = adata.obs[cname].cat.categories
        pal = pd.DataFrame({"palette": c}, index=col)
    except:
        col = adata.obs[cname].cat.categories
        c = c[:len(col)]
        pal = pd.DataFrame({"palette": c}, index=col)
    return pal

import warnings
import matplotlib

def update_color_in_anndata(adata, clustering_name, new_palette):
    """
    Update color information stored in anndata.

    Args:
        adata (anndata._core.anndata.AnnData): anndata object

    """
    if clustering_name not in adata.obs.columns.values:
        raise ValueError(f"{clustering_name} is not found in your anndata.obs")
    if clustering_name not in adata.obs.columns.values:
        raise ValueError(f"{clustering_name}_colors is not found in your anndata.uns \n please make sure you have pre-existing color information for {clusteing_name}")

    old_palette = get_palette(adata=adata, cname=clustering_name)

    new_colors = []
    for cluster in old_palette.index:
        if cluster in new_palette.index.values:
            new_color = new_palette.loc[cluster, "palette"]
        else:
            raise warnings.warn(f"{cluster} not found in the new palette. \nThe color information of {cluster} is not updated.")
            new_color = old_palette.loc[cluster, "palette"]
        new_color = matplotlib.colors.to_hex(new_color)
        new_colors.append(new_color)

    # Update color information
    adata.uns[f"{clustering_name}_colors"] = np.array(new_colors)




@jit(nopython=True)
def _numba_random_seed(value: int) -> None:
    """Same as np.random.seed but for numba"""
    np.random.seed(value)

def _decompose_TFdict(TFdict):
    """
    Args:
        TFdict (dict): Key is target gene, Value is a list of regulatory gene of this target.

    Return:
        (list, list): list of all target gene in the TFdict and list of regulatory gene in the TFdict.
    """

    all_regulatory_genes_in_TFdict = []
    for val in TFdict.values():
        all_regulatory_genes_in_TFdict += list(val)
    all_regulatory_genes_in_TFdict = list(np.unique(all_regulatory_genes_in_TFdict))

    all_target_genes_in_TFdict = list(TFdict.keys())

    return all_target_genes_in_TFdict, all_regulatory_genes_in_TFdict


def _is_perturb_condition_valid(adata, goi, value, safe_range_fold=2):

    """
    Check the input perturb condition is within the safe range.
    Args:
        adata (anndata): scRNA-seq data
        goi (str): Gene of interest
        value (str): Perturb condition input value
        safe_range_fold (float or int): Fold change value.
    Returns:
        Bool

    """
    actual_values = sc.get.obs_df(adata, keys=[goi], layer="imputed_count").values
    min_ = actual_values.min()
    max_ = actual_values.max()
    range_ = max_ - min_

    upper_limit = range_ * (safe_range_fold -1) + max_

    if value <= upper_limit:
        return True
    else:
        return False



#############################################################################
### Sanity check function to evaluate the distribution of simulated value ###
#############################################################################


def _calculate_relative_ratio_of_simulated_value(simulated_count, reference_count):

    """
    CellOracle does not intend to simulate out-of-distribution simulation.
    This function calculate relative value scaled by the reference gene distribution value.


    Args:
        simulated_count (np.nddarray): simulated gene expression matrix.

        reference_count (np.nddarray): unperturbed gene expression matrix.

        pandas.DataFrame: If the value is between 0 to 1, this means the gene value is whitin the reference distribution range.
    """


    results = []
    for gene in reference_count.columns:
        simulated_count[gene]

        relative_ratio = __get_relative_ratio_to_reference_array(
            reference=reference_count[gene],
            query=simulated_count[gene])
        results.append(relative_ratio)

    relative_ratio = pd.concat(results, axis=1)
    relative_ratio.columns = reference_count.columns

    return relative_ratio

def __get_relative_ratio_to_reference_array(reference, query):
    min_ref = np.min(reference)
    max_ref = np.max(reference)
    range_ = max_ref - min_ref
    return (query - min_ref)/range_



# -*- coding: utf-8 -*-
'''

'''


from gimmemotifs.scanner import as_fasta
from tqdm.auto import tqdm

# !!! This function is still under development. May be removed in the future version.
def custom_scan(self, seqs, nreport=100, scan_rc=True, zscore=False, gc=False, batch_size=50000, verbose=True):
    """
    This is the modified version of gimmemotifs's scanner.scan function.

    1) This function uses custom batch size.
    2) Verbose can be silenced.
    """
    seqs = as_fasta(seqs, genome=self.genome)
    if zscore:
        if gc:
            if len(self.meanstd) <= 1:
                self.set_meanstd(gc=gc)
        else:
            if len(self.meanstd) != 1:
                self.set_meanstd(gc=gc)

    # progress bar
    pbar = tqdm(
        desc="scanning",
        unit=" sequences",
        total=len(seqs),
        disable=(verbose==False),  # can be silenced
    )

    #logger.debug("Scanning")
    for batch_idx in range(0, len(seqs), batch_size):
        it = self._scan_sequences(
            seqs.seqs[batch_idx : batch_idx + batch_size],
            nreport,
            scan_rc,
            #zscore=zscore,
        )
        for result in it:
            yield result
            pbar.update(1)
    pbar.close()



# -*- coding: utf-8 -*-
'''
This is a series of custom functions for the inferring of GRN from single cell RNA-seq data.

Codes were written by Kenji Kamimoto.


'''

###########################
### 0. Import libralies ###
###########################


# 0.1. libraries for fundamental data science and data processing

import os

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns

from scipy import stats

from tqdm.notebook import tqdm
import networkx as nx


#import seaborn as sns

settings = {"save_figure_as": "png"}


def transfer_scores_from_links_to_adata(adata, links, method="median"):
    """
    Transfer the summary of network scores (median or mean) per group from Links object into adata.

    Args:
        adata (anndata): anndata

        links (Links): Likns object

        method (str): The method to summarize data.
    """
    cluster_name = links.name
    index_by_cluster = adata.obs[cluster_name].values

    if links.entropy is None:
        links.get_network_entropy()

    if method == "median":
        grouped = links.merged_score.groupby(by="cluster").median()
        grouped_entropy = links.entropy.groupby("cluster").median()
    elif method == "mean":
        grouped = links.merged_score.groupby(by="cluster").mean()
        grouped_entropy = links.entropy.groupby("cluster").mean()


    grouped = grouped.loc[index_by_cluster]
    grouped_entropy= grouped_entropy.loc[index_by_cluster]
    grouped = pd.concat([grouped, grouped_entropy], axis=1)
    grouped.index = adata.obs.index

    adata.obs = pd.concat([adata.obs, grouped], axis=1)
#####################
### Use Network X ###
#####################

def linkList_to_networkgraph(filteredlinkList):
    """
    Convert linkList into Graph object in NetworkX.

    Args:
       filteredlinkList (pandas.DataFrame): GRN saved as linkList.

    Returns:
        Graph object: Network X graph objenct.
    """
    G=nx.DiGraph()
    G_ = nx.from_pandas_edgelist(filteredlinkList, edge_attr=True)
    G.add_edges_from(G_.edges())

    return G


def draw_network(linkList, return_graph=False):
    """
    Plot network graph.

    Args:
       linkList (pandas.DataFrame): GRN saved as linkList.
       return_graph (bool): Whether to return graph object.

    Returns:
        Graph object: Network X graph objenct.
    """
    G = linkList_to_networkgraph(linkList)
    # レイアウトの取得
    pos = nx.spring_layout(G)
    # 可視化
    #nx.draw_networkx_edges(G, pos)
    nx.draw(G, pos,edge_color='black',width=1,linewidths=1,\
node_size=500,node_color='pink',alpha=0.9,arrowstyle='->',
            labels={node:node for node in G.nodes()})
    plt.axis('off')
    plt.show()

    if return_graph:
        return G



# -*- coding: utf-8 -*-



import io
import logging
import os
import pickle
import subprocess
import sys
import math
from copy import deepcopy
import warnings

import pandas as pd
import numpy as np
from tqdm.auto import tqdm
import matplotlib.pyplot as plt
#import h5py
from sklearn.neighbors import KNeighborsRegressor


from ..trajectory.oracle_core import Oracle
from .scatter_to_grid import scatter_value_to_grid_value


from ..utility.hdf5_processing import dump_hdf5, load_hdf5

from ..visualizations.config import CONFIG
from ..visualizations.development_module_visualization import (\
                            plot_cluster_whole,
                            plot_cluster_cells_use,
                            plot_background,
                            plot_pseudotime,
                            plot_reference_flow_on_grid,
                            plot_pseudotime_on_grid)

def load_gradient(file_path):

    """
    Load gradient object saved as hdf5 file.

    Args:
        file_path (str): File path to the hdf5 file.
    """
    obj = load_hdf5(filename=file_path, obj_class=Gradient_calculator, ignore_attrs_if_err=[])


    return obj





class Gradient_calculator():
    def __init__(self, oracle_object=None, adata=None, obsm_key=None, pseudotime_key="Pseudotime", cell_idx_use=None, name=None, gt=None):
        """
        Estimate the direction of differentiation by calculation gradient of pseudotime on the embedding space.
        Please look at web tutorial for example scripts.

        Args:
            adata (anndata): scRNA-seq data in anndata class
            obsm_key (str): Name of dimensional reduction. You can check the list of dimensional reduction data name with "adata.obsm.keys()"
            pseudotime_key (str): Pseudotime data should be stored in adata.obs[pseudotime_key]. Please set the name of pseudotime data in adata.obs
            cluster_column_name (str): If you set cluster_column_name and cluster, you can subset cells to calculate gradient.
                Please look at web tutorial for example codes.
            cluster (str): See above.

        """
        self.cell_idx_use = None
        self.n_neighbors = None
        self.min_mass = None
        self.smooth = None
        self.n_grid = None

        if oracle_object is not None:
            self.load_oracle_object(oracle_object=oracle_object,
                                    cell_idx_use=cell_idx_use,
                                    name=name,
                                    pseudotime_key=pseudotime_key)

        elif adata is not None:
            self.load_adata(adata=adata, obsm_key=obsm_key,
                            pseudotime_key=pseudotime_key,
                            cell_idx_use=cell_idx_use,
                            name=name)

        elif gt is not None:
            self.embedding = gt.embedding.copy()
            self.mass_filter = gt.mass_filter_whole.copy()
            self.mass_filter_whole = gt.mass_filter_whole.copy()
            self.gridpoints_coordinates = gt.gridpoints_coordinates.copy()

            self.n_neighbors = gt.n_neighbors
            self.min_mass = gt.min_mass
            self.smooth = gt.smooth
            self.n_grid = gt.n_grid


    def copy(self):
        """
        Deepcopy itself.
        """
        return deepcopy(self)

    def to_hdf5(self, file_path):
        """
        Save object as hdf5.

        Args:
            file_path (str): file path to save file. Filename needs to end with '.celloracle.oracle'
        """
        if file_path.endswith(".celloracle.gradient"):
            pass
        else:
            raise ValueError("Filename needs to end with '.celloracle.gradient'")

        compression_opts = 5
        dump_hdf5(obj=self, filename=file_path,
                  data_compression=compression_opts,  chunks=(2048, 2048),
                  noarray_compression=compression_opts, pickle_protocol=4)

    def load_adata(self, adata, obsm_key, cell_idx_use=None, name=None, pseudotime_key="Pseudotime"):

        self.name = name
        self.embedding = adata.obsm[obsm_key].copy()
        self.pseudotime = adata.obs[pseudotime_key].values.copy()

        if cell_idx_use is not None:
            self.cell_idx_use = np.array(cell_idx_use)

    def load_oracle_object(self, oracle_object, cell_idx_use=None, name=None, pseudotime_key="Pseudotime"):
        self.load_adata(adata=oracle_object.adata,
                        obsm_key=oracle_object.embedding_name,
                        cell_idx_use=cell_idx_use,
                        name=name,
                        pseudotime_key=pseudotime_key)

    def calculate_p_mass(self, smooth=0.8, n_grid=40, n_neighbors=200, n_jobs=-1):

        x_min, y_min = self.embedding.min(axis=0)
        x_max, y_max = self.embedding.max(axis=0)
        xylim = ((x_min, x_max), (y_min, y_max))
        steps = (n_grid, n_grid)

        if self.cell_idx_use is None:

            total_p_mass, gridpoints_coordinates = calculate_p_mass(self.embedding, smooth=smooth, steps=steps,
                                  n_neighbors=n_neighbors, n_jobs=n_jobs, xylim=xylim)
            total_p_mass_whole = total_p_mass.copy()

        else:
            total_p_mass, gridpoints_coordinates = calculate_p_mass(self.embedding[self.cell_idx_use, :], smooth=smooth, steps=steps,
                                  n_neighbors=n_neighbors, n_jobs=n_jobs, xylim=xylim)

            total_p_mass_whole, _ = calculate_p_mass(self.embedding, smooth=smooth, steps=steps,
                                  n_neighbors=n_neighbors, n_jobs=n_jobs, xylim=xylim)

        self.n_neighbors = n_neighbors
        self.smooth = smooth
        self.n_grid = n_grid

        self.total_p_mass = total_p_mass
        self.total_p_mass_whole = total_p_mass_whole
        self.gridpoints_coordinates = gridpoints_coordinates


    def suggest_mass_thresholds(self, n_suggestion=12, s=1, n_col=4):

        min_ = self.total_p_mass.min()
        max_ = self.total_p_mass.max()
        suggestions = np.linspace(min_, max_/2, n_suggestion)

        n_rows = math.ceil(n_suggestion / n_col)

        fig, ax = plt.subplots(n_rows, n_col, figsize=[5*n_col, 5*n_rows])
        if n_rows == 1:
            ax = ax.reshape(1, -1)

        row = 0
        col = 0
        for i in range(n_suggestion):

            ax_ = ax[row, col]

            col += 1
            if col == n_col:
                col = 0
                row += 1

            idx = self.total_p_mass > suggestions[i]

                #ax_.scatter(gridpoints_coordinates[mass_filter, 0], gridpoints_coordinates[mass_filter, 1], s=0)
            ax_.scatter(self.embedding[:, 0], self.embedding[:, 1], c="lightgray", s=s)
            ax_.scatter(self.gridpoints_coordinates[idx, 0],
                       self.gridpoints_coordinates[idx, 1],
                       c="black", s=s)
            ax_.set_title(f"min_mass: {suggestions[i]: .2g}")
            ax_.axis("off")


    def calculate_mass_filter(self, min_mass=0.01, plot=False):

        self.min_mass = min_mass
        self.mass_filter = (self.total_p_mass < min_mass)
        self.mass_filter_whole = (self.total_p_mass_whole < min_mass)

        if plot:
            fig, ax = plt.subplots(figsize=[5,5])

            #ax_.scatter(gridpoints_coordinates[mass_filter, 0], gridpoints_coordinates[mass_filter, 1], s=0)
            ax.scatter(self.embedding[:, 0], self.embedding[:, 1], c="lightgray", s=10)
            ax.scatter(self.gridpoints_coordinates[~self.mass_filter, 0],
                       self.gridpoints_coordinates[~self.mass_filter, 1],
                       c="black", s=0.5)
            ax.set_title("Grid points selected")
            ax.axis("off")

    def transfer_data_into_grid(self, args={}, plot=False):

        if not args:
            args = {"method": "knn",
                    "n_knn": 30}

        # Prepare input data_new
        if self.cell_idx_use is None:
            embedding = self.embedding
            grid = self.gridpoints_coordinates
            value = self.pseudotime
        else:
            embedding = self.embedding[self.cell_idx_use, :]
            grid = self.gridpoints_coordinates
            value = self.pseudotime[self.cell_idx_use]

        # Remove inf
        if np.inf in value:
            # Clip inf
            warnings.warn("Inf value found in the pseudotime data. The inf value is replaced with non-inf max value.", UserWarning)
            _clip_inf_value(data=value)

        # Data calculation for each grid point
        self.pseudotime_on_grid = scatter_value_to_grid_value(embedding=embedding,
                                                              grid=grid,
                                                              value=value,
                                                              **args)

        if plot:
            fig, ax = plt.subplots(1, 2, figsize=[10,5])

            s = 10
            s_grid = 20
            show_background = True
            ##
            ax_ = ax[0]
            plot_pseudotime(self, ax=ax_, s=s, show_background=show_background)
            ax_.set_title("Pseudotime")


            ####
            ax_ = ax[1]
            plot_pseudotime_on_grid(self, ax=ax_, s=s_grid, show_background=show_background)
            ax_.set_title("Pseudotime on grid")




    def calculate_gradient(self, scale_factor="l2_norm_mean", normalization="sqrt"):

        # Gradient calculation
        gradient = get_gradient(value_on_grid=self.pseudotime_on_grid.copy())

        if normalization == "sqrt":
            gradient = normalize_gradient(gradient, method="sqrt")

        if scale_factor == "l2_norm_mean":
            # divide gradient by the mean of l2 norm.
            l2_norm = np.linalg.norm(gradient, ord=2, axis=1)
            scale_factor = 1 / l2_norm.mean()

        self.ref_flow = gradient * scale_factor


    def plot_dev_flow_on_grid(self, ax=None, scale=CONFIG["scale_dev"], show_background=True, s=CONFIG["s_scatter"], args={}):
        plot_reference_flow_on_grid(self, ax=ax, scale=scale, show_background=show_background, s=s, args=args)

    def plot_reference_flow_on_grid(self, ax=None, scale=CONFIG["scale_dev"], show_background=True, s=CONFIG["s_scatter"], args={}):
        plot_reference_flow_on_grid(self, ax=ax, scale=scale, show_background=show_background, s=s, args=args)

    def visualize_results(self, scale=30, s=1, s_grid=30, show_background=True):


        fig, ax = plt.subplots(1, 5, figsize=[25,5])

        ##
        ax_ = ax[0]
        plot_pseudotime(self, ax=ax_, s=s, show_background=show_background)
        ax_.set_title("Pseudotime")

        ####
        ax_ = ax[1]
        plot_pseudotime_on_grid(self, ax=ax_, s=s_grid, show_background=show_background)
        ax_.set_title("Pseudotime on grid")

        ###
        ax_ = ax[2]
        plot_pseudotime_on_grid(self, ax=ax_, s=s_grid, show_background=show_background)
        plot_reference_flow_on_grid(self, ax=ax_, scale=scale, show_background=False, s=s)
        ax_.set_title("Gradient of pseudotime \n(Development flow)")

        ###
        ax_ = ax[3]
        plot_reference_flow_on_grid(self, ax=ax_, scale=scale, show_background=show_background, s=s)
        ax_.set_title("Gradient of pseudotime \n(=Development flow)")

        ####
        ax_ = ax[4]
        plot_pseudotime(self, ax=ax_, s=s, show_background=show_background)
        plot_reference_flow_on_grid(self, ax=ax_, scale=scale, show_background=False, s=s)
        ax_.set_title("Pseudotime + \nDevelopment flow")




    def plot_pseudotime(self, ax=None, s=CONFIG["s_scatter"],show_background=True, args=CONFIG["default_args"], cmap="rainbow"):
        plot_pseudotime(self, ax=None, s=s, show_background=show_background, cmap=cmap, args=args)

def aggregate_Gradient_objects(gradient_object_list, base_gt=None, fill_na=True):

    # Aggregate pseudotime
    pseudotime_stack = [i.pseudotime for i in gradient_object_list]
    new_pseudotime, _ = _aggregate_array(value_stack=pseudotime_stack,
                                         filter_stack=[np.isnan(i) for i in pseudotime_stack])

    # Aggregate "pseudotime on grid" and "gradient"
    mass_filter_stack = [i.mass_filter for i in gradient_object_list]

    pseudotime_grid_stack = [i.pseudotime_on_grid for i in gradient_object_list]
    new_pseudotime_grid, _ = _aggregate_array(value_stack=pseudotime_grid_stack,
                                              filter_stack=mass_filter_stack)

    gradient_stack = [i.ref_flow for i in gradient_object_list]
    new_gradient, new_mass_filter  = _aggregate_array(value_stack=gradient_stack,
                                                      filter_stack=mass_filter_stack )

    if base_gt is None:
        gt = Gradient_calculator(gt=gradient_object_list[0])
        gt.pseudotime_on_grid = new_pseudotime_grid
        gt.ref_flow = new_gradient

    else:
        gt = base_gt
        gt.pseudotime_on_grid[~new_mass_filter] = new_pseudotime_grid[~new_mass_filter]
        gt.ref_flow[~new_mass_filter, :] = new_gradient[~new_mass_filter, :]

    if fill_na:
        y_with_nan= gt.pseudotime_on_grid.copy()
        y_with_nan[(new_mass_filter & (~gt.mass_filter))] = np.nan
        y_with_nan = pd.Series(y_with_nan)
        # Replace na and inf with NN's value
        y_filled =  _fill_inf_and_na(X=gt.gridpoints_coordinates[~gt.mass_filter, :],
                            y_with_inf=y_with_nan[~gt.mass_filter])

        gt.pseudotime_on_grid[~gt.mass_filter] = y_filled

    gt.pseudotime = new_pseudotime

    return gt



def _aggregate_array(value_stack, filter_stack):

    new_value = np.zeros_like(value_stack[0])
    counts = np.zeros_like(value_stack[0])
    for fil, val in zip(filter_stack, value_stack):
        if len(val.shape) == 1:
            new_value[~fil] += val[~fil]
            counts[~fil] +=1
        elif len(val.shape) == 2:
            new_value[~fil, :] += val[~fil, :]
            counts[~fil, :] +=1

    new_value[counts != 0] /= counts[counts != 0]

    new_filter = (counts.sum(axis=-1) == 0)

    return new_value, new_filter

def normalize_gradient(gradient, method="sqrt"):
    """
    Normalize length of 2D vector
    """

    if method == "sqrt":

        size = np.sqrt(np.power(gradient, 2).sum(axis=1))
        size_sq = np.sqrt(size)
        size_sq[size_sq == 0] = 1
        factor = np.repeat(np.expand_dims(size_sq, axis=1), 2, axis=1)

    return gradient / factor

from scipy.stats import norm as normal
from sklearn.neighbors import NearestNeighbors


def calculate_p_mass(embedding, smooth=0.5, steps=(40, 40),
                          n_neighbors=100, n_jobs=4, xylim=((None, None), (None, None))):
    """Calculate the velocity using a points on a regular grid and a gaussian kernel

    Note: the function should work also for n-dimensional grid

    Arguments
    ---------
    embedding:

    smooth: float, smooth=0.5
        Higher value correspond to taking in consideration further points
        the standard deviation of the gaussian kernel is smooth * stepsize
    steps: tuple, default
        the number of steps in the grid for each axis
    n_neighbors:
        number of neighbors to use in the calculation, bigger number should not change too much the results..
        ...as soon as smooth is small
        Higher value correspond to slower execution time
    n_jobs:
        number of processes for parallel computing
    xymin:
        ((xmin, xmax), (ymin, ymax))

    Returns
    -------
    total_p_mass: np.ndarray
        density at each point of the grid

    """

    # Prepare the grid
    grs = []
    for dim_i in range(embedding.shape[1]):
        m, M = np.min(embedding[:, dim_i]), np.max(embedding[:, dim_i])

        if xylim[dim_i][0] is not None:
            m = xylim[dim_i][0]
        if xylim[dim_i][1] is not None:
            M = xylim[dim_i][1]

        m = m - 0.025 * np.abs(M - m)
        M = M + 0.025 * np.abs(M - m)
        gr = np.linspace(m, M, steps[dim_i])
        grs.append(gr)

    meshes_tuple = np.meshgrid(*grs)
    gridpoints_coordinates = np.vstack([i.flat for i in meshes_tuple]).T

    nn = NearestNeighbors(n_neighbors=n_neighbors, n_jobs=n_jobs)
    nn.fit(embedding)
    dists, neighs = nn.kneighbors(gridpoints_coordinates)

    std = np.mean([(g[1] - g[0]) for g in grs])
    # isotropic gaussian kernel
    gaussian_w = normal.pdf(loc=0, scale=smooth * std, x=dists)
    total_p_mass = gaussian_w.sum(1)
    gridpoints_coordinates

    return total_p_mass, gridpoints_coordinates

def get_gradient(value_on_grid):
    # Gradient calculation
    n = int(np.sqrt(value_on_grid.shape[0]))
    value_on_grid_as_matrix = value_on_grid.reshape(n, n)
    dy, dx = np.gradient(value_on_grid_as_matrix)
    gradient = np.stack([dx.flatten(), dy.flatten()], axis=1)

    return gradient


def _fill_inf_and_na(X, y_with_inf):
    # Make bool list for the cells to be replaced
    na_cell_bool = y_with_inf.replace(np.inf, np.nan).isna().values

    # Make KNN model and fitting, prediction
    knn = KNeighborsRegressor(n_neighbors=30)
    knn.fit(X[~na_cell_bool, :], y_with_inf[~na_cell_bool])
    y_filled = y_with_inf.copy()
    y_filled[na_cell_bool] = knn.predict(X[na_cell_bool, :])

    # check
    assert(y_filled.replace(np.inf, np.nan).isna().sum() == 0)

    return y_filled

def _clip_inf_value(data):
    """
    This function replace inf with non-inf max value
    """
    max_without_inf = data[data != np.inf].max()
    data[data == np.inf] = max_without_inf

    #print(max_without_inf)



# -*- coding: utf-8 -*-



import io
import logging
import os
import pickle
import subprocess
import sys
from copy import deepcopy

import pandas as pd
import numpy as np
from tqdm.auto import tqdm
import matplotlib.pyplot as plt
#import h5py


from IPython.display import display, HTML
from ipywidgets import interactive, FloatText

from .development_module import Oracle_development_module
from ..visualizations.config import CONFIG


class Oracle_systematic_analysis_helper(Oracle_development_module):

    def __init__(self, hdf5_file_path, memory_save_mode=False):
        super().__init__()

        self.set_hdf_path(path=hdf5_file_path)
        self.hdf5_info = self.get_hdf5_info()
        self.estimated_scales_for_visulization = None
        self.negative_ip_sum = None
        self.positive_ip_sum = None
        self.memory_save_mode = memory_save_mode
        self._exemptions_when_del_attrs = \
            ["hdf5_info", "negative_ip_sum", "positive_ip_sum", "estimated_scales_for_visulization"]

    def get_negative_ip_sum_for_all_data(self, verbose=True, return_result=True):

        self.del_attrs()

        gene_misc_lists = self.hdf5_info["gene_misc_lists"]

        li = []
        if verbose:
            loop = tqdm(gene_misc_lists)
        else:
            loop = gene_misc_lists

        for gene, misc in loop:

            self.load_hdf5(gene=gene, misc=misc, specify_attributes=["inner_product_df"])
            df = self.get_sum_of_negative_ips()
            df["gene"] = gene
            df["misc"] = misc
            li.append(df)

        negative_ip_sum = pd.concat(li, axis=0).reset_index(drop=True)

        self.negative_ip_sum = negative_ip_sum

        # Clear memory
        self.del_attrs()

        if return_result:
            return self.negative_ip_sum


    def get_positive_ip_sum_for_all_data(self, verbose=True, return_result=True):

        self.del_attrs()

        gene_misc_lists = self.hdf5_info["gene_misc_lists"]

        li = []
        if verbose:
            loop = tqdm(gene_misc_lists)
        else:
            loop = gene_misc_lists

        self.del_attrs()

        for gene, misc in loop:
            self.load_hdf5(gene=gene, misc=misc, specify_attributes=["inner_product_df"])
            df = self.get_sum_of_positive_ips()
            df["gene"] = gene
            df["misc"] = misc
            li.append(df)

        positive_ip_sum = pd.concat(li, axis=0).reset_index(drop=True)

        self.positive_ip_sum = positive_ip_sum

        # Clear memory
        self.del_attrs()

        if return_result:
            return positive_ip_sum

    def sort_TFs_by_neagative_ip(self, misc, pseudotime="0,1,2,3,4,5,6,7,8,9"):
        if self.negative_ip_sum is None:
            self.get_negative_ip_sum_for_all_data(return_result=False, verbose=False)

        negative_ip_sum = self.negative_ip_sum

        # Focus on a misc
        df = negative_ip_sum[negative_ip_sum.misc == misc]

        # Focus on a specific pseudotime
        pseudotime = [i for i in pseudotime if i in list("0123456789")]
        df = df[df.pseudotime_id.isin(pseudotime)]

        # Get sum of negative ip values
        df = df[["gene", "score"]].groupby(["gene"]).sum()

        # Sore by sum score
        df = df.sort_values("score", ascending=True).reset_index(drop=False)

        return df

    def sort_TFs_by_positive_ip(self, misc, pseudotime="0,1,2,3,4,5,6,7,8,9"):
        if self.positive_ip_sum is None:
            self.get_positive_ip_sum_for_all_data(return_result=False, verbose=False)

        positive_ip_sum = self.positive_ip_sum

        # Focus on a misc
        df = positive_ip_sum[positive_ip_sum.misc == misc]

        # Focus on a specific pseudotime
        pseudotime = [i for i in pseudotime if i in list("0123456789")]
        df = df[df.pseudotime_id.isin(pseudotime)]

        # Get sum of negative ip values
        df = df[["gene", "score"]].groupby(["gene"]).sum()

        # Sore by sum score
        df = df.sort_values("score", ascending=False).reset_index(drop=False)

        return df

    def interactive_sort_TFs_by_neagative_ip(self):

        if self.hdf5_info is None:
            self.get_hdf5_info()

        def wrapper(misc, pseudotime, n_TFs=20):
            df = self.sort_TFs_by_neagative_ip(misc=misc, pseudotime=pseudotime)

            print(f"Top {n_TFs} in {misc}")
            display(HTML(df.iloc[:min(n_TFs, df.shape[0])].to_html()))

        interactive_table = interactive(wrapper,
                               #{'manual': True},
                               misc=self.hdf5_info["misc_list"],
                               pseudotime="0,1,2,3,4,5,6,7,8,9",
                               n_TFs=(5, 50, 1))

        return interactive_table

    def interactive_sort_TFs_by_positive_ip(self):

        if self.hdf5_info is None:
            self.get_hdf5_info()

        def wrapper(misc, pseudotime, n_TFs=20):
            df = self.sort_TFs_by_positive_ip(misc=misc, pseudotime=pseudotime)

            print(f"Top {n_TFs} in {misc}")
            display(HTML(df.iloc[:min(n_TFs, df.shape[0])].to_html()))

        interactive_table = interactive(wrapper,
                               #{'manual': True},
                               misc=self.hdf5_info["misc_list"],
                               pseudotime="0,1,2,3,4,5,6,7,8,9",
                               n_TFs=(5, 50, 1))

        return interactive_table

    def _interactive_calculate_negative_ps_p_value(self):
        """
        8/21/2021. This function is under development
        Get p-values for negative PS sum score by comparing nPS to randomized GRN nPS.
        """

        if self.hdf5_info is None:
            self.get_hdf5_info()

        def wrapper(misc, pseudotime, n_TFs=20):
            df = self._calculate_negative_ps_p_value(misc=misc, pseudotime=pseudotime)


            print(f"Top {n_TFs} in {misc}")
            display(HTML(df.iloc[:min(n_TFs, df.shape[0])].to_html()))

        interactive_table = interactive(wrapper,
                               #{'manual': True},
                               misc=self.hdf5_info["misc_list"],
                               pseudotime="0,1,2,3,4,5,6,7,8,9",
                               n_TFs=(5, 50, 1))

        return interactive_table

    def _interactive_calculate_positive_ps_p_value(self):
        """
        8/21/2021. This function is under development
        Get p-values for negative PS sum score by comparing nPS to randomized GRN nPS.
        """

        if self.hdf5_info is None:
            self.get_hdf5_info()

        def wrapper(misc, pseudotime, n_TFs=20):
            df = self._calculate_positive_ps_p_value(misc=misc, pseudotime=pseudotime)


            print(f"Top {n_TFs} in {misc}")
            display(HTML(df.iloc[:min(n_TFs, df.shape[0])].to_html()))

        interactive_table = interactive(wrapper,
                               #{'manual': True},
                               misc=self.hdf5_info["misc_list"],
                               pseudotime="0,1,2,3,4,5,6,7,8,9",
                               n_TFs=(5, 50, 1))

        return interactive_table


    def _update_inner_product_df(self, n_bins=10, verbose=True):
        self.del_attrs()

        gene_misc_lists = self.hdf5_info["gene_misc_lists"]

        li = []
        if verbose:
            loop = tqdm(gene_misc_lists)
        else:
            loop = gene_misc_lists

        self.del_attrs()

        for gene, misc in loop:
            self.load_hdf5(gene=gene, misc=misc)
            self.calculate_inner_product()
            self.calculate_digitized_ip(n_bins=n_bins)
            self.dump_hdf5(gene=gene, misc=misc)
            # Clear memory
            self.del_attrs()


    def _calculate_negative_ps_p_value(self, misc, pseudotime="0,1,2,3,4,5,6,7,8,9", verbose=True):

        """
        8/21/2021. This function is under development
        Get p-values for negative PS sum score by comparing nPS to randomized GRN nPS.
        """

        self.del_attrs()

        gene_lists = self.hdf5_info["gene_list"]

        p_list = []
        ps_sums = []
        ps_sum_randoms = []
        if verbose:
            loop = tqdm(gene_lists)
        else:
            loop = gene_lists

        for gene in loop:

            self.load_hdf5(gene=gene, misc=misc, specify_attributes=["inner_product_df"])
            if "score_randomized" not in self.inner_product_df.columns:
                raise ValueError("please update inner_product_df first")
            p, ps_sum, ps_sum_random = self.get_negative_PS_p_value(pseudotime=pseudotime, return_ps_sum=True, plot=False)
            p_list.append(p)
            ps_sum_randoms.append(ps_sum_random)
            ps_sums.append(ps_sum)

        # Clear memory
        self.del_attrs()

        # p-value correction
        p_corrected = np.clip(np.array(p_list)*len(gene_lists), 0, 1)

        result = pd.DataFrame({"gene": gene_lists, "ps_sum":ps_sums, "ps_sum_random": ps_sum_randoms,
                               "p": p_list, "p_adj": p_corrected})

        result = result.sort_values("ps_sum", ascending=False).reset_index(drop=True)

        return result

    def _calculate_positive_ps_p_value(self, misc, pseudotime="0,1,2,3,4,5,6,7,8,9", verbose=True):

        """
        8/21/2021. This function is under development
        Get p-values for negative PS sum score by comparing nPS to randomized GRN nPS.
        """

        self.del_attrs()

        gene_lists = self.hdf5_info["gene_list"]

        p_list = []
        ps_sums = []
        ps_sum_randoms = []
        if verbose:
            loop = tqdm(gene_lists)
        else:
            loop = gene_lists

        for gene in loop:

            self.load_hdf5(gene=gene, misc=misc, specify_attributes=["inner_product_df"])
            if "score_randomized" not in self.inner_product_df.columns:
                raise ValueError("please update inner_product_df first")
            p, ps_sum, ps_sum_random = self.get_positive_PS_p_value(pseudotime=pseudotime, return_ps_sum=True, plot=False)
            p_list.append(p)
            ps_sum_randoms.append(ps_sum_random)
            ps_sums.append(ps_sum)

        # Clear memory
        self.del_attrs()

        # p-value correction
        p_corrected = np.clip(np.array(p_list)*len(gene_lists), 0, 1)

        result = pd.DataFrame({"gene": gene_lists, "ps_sum":ps_sums, "ps_sum_random": ps_sum_randoms,
                               "p": p_list, "p_adj": p_corrected})

        result = result.sort_values("ps_sum", ascending=False).reset_index(drop=True)

        return result


    def estimate_scale_for_visualization(self, return_result=True):

        if self.hdf5_info is None:
            self.get_hdf5_info()

        misc_list = self.hdf5_info["misc_list"]

        scales = np.array([self._estimate_scale(misc=misc) for misc in misc_list])
        scales = pd.DataFrame(scales, index=misc_list, columns=["vm", "simulation", "pseudotime"])

        self.estimated_scales_for_visulization = scales

        if return_result:
            return scales


    def _estimate_scale(self, misc):

        self.del_attrs()

        reference_scaling_dictionary = {"inner_product": 0.5,
                                        "flow": 5,
                                        "ref_flow": 18}

        scales = []
        for i in ["inner_product", "flow", "ref_flow"]:
            max_ = self.__check_max_of_max_norm_attr(misc=misc, attr_name=i)
            scales.append(max_ * reference_scaling_dictionary[i])

        # Clear memory
        self.del_attrs()

        # this is the lis of scales
        return scales

    def __check_max_of_max_norm_attr(self, misc, attr_name):

        """
        (1) Load attr and check max norm value.
        (2) Repeat step (1) for all genes in the misc.
        (3) Return max of all max of norm value list.

        if attr is 2D value, calculate max of l2 norm.

        """

        if self.hdf5_info is None:
            self.get_hdf5_info()

        gene_list = self.hdf5_info["misc_gene_dictionary"][misc]


        if attr_name == "ref_flow": # Don't need to check all gene because ref_flow is unique to misc.
            max_ = self.__check_max_norm_attr(misc=misc, gene=gene_list[0], attr_name=attr_name)

        else:
            maxs = []
            for gene in gene_list:
                abs_max = self.__check_max_norm_attr(misc=misc, gene=gene, attr_name=attr_name)
                maxs.append(abs_max)
            max_ = max(maxs)

        return max_

    ### For interactive systematic visualization
    def __check_max_norm_attr(self, misc, gene, attr_name):

        """
        Load attr and check max norm value.
        if attr is 2D value, calculate max of l2 norm.

        """
        self.load_hdf5(gene=gene, misc=misc, specify_attributes=[attr_name])
        attr = getattr(self, attr_name)

        if attr.ndim == 1:
            abs_max = np.absolute(attr).max()
        elif attr.ndim == 2:
            abs_max = np.linalg.norm(attr, axis=1).max()

        return abs_max

    def interactive_visualize_layout_0(self, scale_simulation=None, scale_pseudotime=None, vm=None, s=5, s_grid=30):

        # 1. Define a custom function
        def wrapper(gene, misc, scale_simulation, scale_pseudotime, vm, background):
            # Load data
            #self.del_attrs()

            self.load_hdf5(gene=gene, misc=misc)
            # Visualize
            self.visualize_development_module_layout_0(s=s,
                                                       scale_for_simulation=scale_simulation,
                                                       s_grid=s_grid,
                                                       scale_for_pseudotime=scale_pseudotime,
                                                       vm=vm,
                                                       show_background=background)
            print("Gene: ", gene)
            # Delete loaded data
            #self.del_attrs(exemptions=self._exemptions_when_del_attrs)


        # 2. Parameter setting
        # Use automatically estimated value for default scale parameters if it is not specified.

        if np.any([i is None for i in [scale_simulation, scale_pseudotime, vm]]):
            if self.estimated_scales_for_visulization is None:
                self.estimate_scale_for_visualization(return_result=False)
            scale = self.estimated_scales_for_visulization.max(axis=0)
            scale_simulation, scale_pseudotime, vm = scale[["simulation", "pseudotime", "vm"]]
            #scale_simulation, vm = [i * 0.5 for i in [scale_simulation, vm]]
            scale_simulation, scale_pseudotime, vm = [np.round(i, 2) for i in [scale_simulation, scale_pseudotime, vm]]

        ui_gene_list = self.hdf5_info["gene_list"]
        ui_misc_list = self.hdf5_info["misc_list"]

        interactive_viz = interactive(wrapper,
                                       #{'manual': True},
                                  scale_simulation=FloatText(value=scale_simulation),
                                  scale_pseudotime=FloatText(value=scale_pseudotime),
                                  vm=FloatText(value=vm),
                                  gene=ui_gene_list,
                                  misc=ui_misc_list,
                                  background=[True, False])

        return interactive_viz


    def get_all_ips(self, misc):

        misc = "Whole_cells"
        genes = self.hdf5_info["misc_gene_dictionary"][misc]

        # Load inner product for all genes
        ips = []
        for gene in genes:
            self.load_hdf5(gene=gene, misc=misc, specify_attributes=["inner_product"])
            ips.append(helper.inner_product)
        ips = np.stack(ips)
        ips = pd.DataFrame(ips, index=genes)

        # Clear memory
        self.del_attrs()

        return ips

    def get_corrcoef_ip(self, misc):

        ips = self.get_all_ips(misc=misc)

        # Calculate correlation
        corrcoef = np.corrcoef(ips)

        return corrcoef



# -*- coding: utf-8 -*-

import os
import sys

import pandas as pd
import numpy as np
import matplotlib.pyplot as plt

import scanpy as sc
from sklearn.neighbors import KNeighborsRegressor

from ..trajectory.oracle_utility import _adata_to_color_dict

class Pseudotime_calculator():
    """
    This is a custom object for pseudotime calculation.

    """
    def __init__(self, oracle_object=None, adata=None, obsm_key=None, cluster_column_name=None, lineage_dictionary=None):
        self.adata = None
        self.adata_raw = None
        self.obsm_key = None
        self.cluster_column_name = None
        self.lineage_dictionary = None

        if adata is not None:
            if (obsm_key is not None) & (cluster_column_name is not None):
                self.load_anndata(adata=adata, obsm_key=obsm_key, cluster_column_name=cluster_column_name)
            else:
                print("Please set obsm_key and cluster_column_name to instantiate Pseudotime_calculator with anndata.")

        if oracle_object is not None:
            self.load_oracle_object(oracle_object=oracle_object)

        if lineage_dictionary is not None:
            self.set_lineage(lineage_dictionary=lineage_dictionary)

    def load_anndata(self, adata, obsm_key, cluster_column_name):
        self.adata = adata.copy()
        self.adata_raw = adata.copy()
        self.obsm_key = obsm_key
        self.cluster_column_name = cluster_column_name
        self.cluster_list = sorted(list(adata.obs[cluster_column_name].unique()))

    def load_oracle_object(self, oracle_object):
        self.load_anndata(adata=oracle_object.adata,
                          obsm_key=oracle_object.embedding_name,
                          cluster_column_name=oracle_object.cluster_column_name)

    def reset(self):
        self.adata = self.adata_raw.copy()

    def set_lineage(self, lineage_dictionary):
        self.lineage_dictionary = lineage_dictionary.copy()

        for lineage, clusters in lineage_dictionary.items():
            self.adata.obs[lineage] = self.adata.obs[self.cluster_column_name].isin(clusters)
            self.adata.obs[lineage] = self.adata.obs[lineage].astype("str").astype("category")
            self.adata.uns[f"{lineage}_colors"] = ["#D0D3D4", "#EC7063"]

    def re_calculate_diffmap(self, n_neighbors=50, use_rep=None):
        del self.adata.obsm["X_diffmap"]
        if use_rep is not None:
            use_rep = self.obsm_key
        sc.pp.neighbors(self.adata, n_neighbors=50, use_rep=use_rep)
        sc.tl.diffmap(self.adata)

    def set_root_cells(self, root_cells):
        """
        Args:
            root_cells (dictionary): key is the name of lineage, the value is name of cell
        """
        self.root_cells = root_cells.copy()

        # check if the format is correct
        for lineage, root_cell in root_cells.items():
            if lineage not in self.adata.obs.columns:
                print(f"The key, {lineage} is not in the lineage data. Make sure the root_cells has correct format.")
            if root_cell not in self.adata.obs.index[self.adata.obs[lineage] == "True"]:
                print(f"The cell, {root_cell} in the {lineage} was not found in your adata. Please make sure the format is correct.")

    def get_pseudotime_per_each_lineage(self, root_cells=None, plot=False):

        if root_cells is not None:
            self.set_root_cells(root_cells=root_cells)

        pseudotime_estimation_for_each_lineage(adata=self.adata, root_cells=self.root_cells)

        # Impute inf or nan value with real value
        main_pseudotime = self.adata.obs["Pseudotime"]
        main_pseudotime = main_pseudotime.replace(np.inf, np.nan)

        if (main_pseudotime.isna().sum() > 0):
            if main_pseudotime.isna().mean() >= 0.5:
                raise ValueError("Found too many cells that failed to get pseudotime. Please check lineage dictionary.")
            else:
                # Replace na and inf with NN's value
                y_filled =  _fill_inf_and_na(X=self.adata.obsm[self.obsm_key],
                                    y_with_inf=self.adata.obs["Pseudotime"])
                self.adata.obs["Pseudotime"] = y_filled

        if plot:
            #sc.pl.umap(self.adata, color=[i for i in self.adata.obs.columns if "Pseudotime" in i], ncols=3, size=20, cmap="rainbow")
            self.plot_pseudotime()

    def plot_cluster(self, s=10, fontsize=8):

        embedding = self.adata.obsm[self.obsm_key]

        color_dict = _adata_to_color_dict(adata=self.adata, cluster_use=self.cluster_column_name)

        fig = plt.figure()
        for cluster, color in color_dict.items():
            idx = np.where(self.adata.obs[self.cluster_column_name] == cluster)[0]
            plt.scatter(embedding[idx, 0], embedding[idx, 1], color=color, s=s, label=cluster)

        plt.legend(bbox_to_anchor=(1.05, 0), loc='lower left', borderaxespad=0, fontsize=fontsize)
        plt.title(self.cluster_column_name)
        plt.axis("off")
        plt.show()

    def plot_lineages(self, s=10):

        embedding = self.adata.obsm[self.obsm_key]

        for lineage in self.lineage_dictionary.keys():
            color = ["#EC7063" if i=="True" else "#D0D3D4" for i in self.adata.obs[lineage].values]

            fig = plt.figure()
            plt.scatter(embedding[:, 0], embedding[:, 1], color=color, s=s)
            plt.title(lineage)
            plt.axis("off")
            plt.show()


    def plot_root_cells(self, s=10):

        embedding = self.adata.obsm[self.obsm_key]

        for lineage, root_cell in self.root_cells.items():
            color = ["#EC7063" if i=="True" else "#D0D3D4" for i in self.adata.obs[lineage].values]
            root_cell_idx = np.where(self.adata.obs.index == root_cell)[0]
            fig = plt.figure()
            plt.scatter(embedding[:, 0], embedding[:, 1], color=color, s=s)
            plt.scatter(embedding[root_cell_idx, 0], embedding[root_cell_idx, 1], color="black", s=s*10, label="root_cell")
            plt.legend()
            plt.title(lineage)
            plt.axis("off")
            plt.show()


    def plot_pseudotime(self, s=10, cmap="rainbow"):

        embedding = self.adata.obsm[self.obsm_key]

        pseudotime_list = [f"Pseudotime_{i}" for i in self.root_cells.keys()] + ["Pseudotime"]
        for pseudotime in pseudotime_list:

            color = self.adata.obs[pseudotime]
            idx = np.where(~color.isna())[0]

            fig = plt.figure()
            plt.scatter(embedding[idx, 0], embedding[idx, 1], c=color[idx], s=s, cmap=cmap)
            plt.title(pseudotime)
            plt.axis("off")
            plt.show()


def pseudotime_estimation_for_each_lineage(adata, root_cells):
    """
    Calculate pseudotime using dpt method in scanpy. https://scanpy.readthedocs.io/en/stable/api/scanpy.tl.dpt.html#scanpy.tl.dpt
    This function consists of four steps below.

    (1) Split scRNA-seq data into subgroup (lineage).
    (2) Calculate pseudotime by dpt
    (3) Re-aggregate cells into original adata
    (4) Fill inf and nan value in the pseudotime by KNN regression.

    Args:
        adata (anndata): scRNA-seq data
        root_cells (dictionary): The key is the name of lineage. This information should exist in the adata.obs
            The value of this dictionary is the name of root cell.

    Returns:
        None: Results will be stored in the adata.obs: (i)The pseudotime for each lineage and (ii)the pseudotime for whole population.

    """

    pseudotime_columns = []
    for lineage, root_cell in root_cells.items():
        # Make subset of adata for single lineage
        bool_ = adata.obs[lineage].values

        if isinstance(bool_[0], str):
            bool_ = [{"True": True, "False": False}[i] for i in bool_]
        cells_in_the_lineage = adata.obs.index[bool_].values
        adata_ = adata[cells_in_the_lineage, :]
        # Set root
        adata_.uns['iroot'] = np.where(adata_.obs.index == root_cell)[0][0]

        # Calculate pseudotime
        sc.tl.dpt(adata_)

        # Save calculated data into original adata
        adata.obs[f'Pseudotime_{lineage}'] = np.nan
        adata.obs.loc[adata_.obs.index.values, f'Pseudotime_{lineage}'] = adata_.obs["dpt_pseudotime"]

        pseudotime_columns.append(f'Pseudotime_{lineage}')

    # Calculate pseudotime average
    adata.obs["Pseudotime"] = adata.obs[pseudotime_columns].mean(axis=1, skipna=True)



def _fill_inf_and_na(X, y_with_inf):
    # Make bool list for the cells to be replaced
    na_cell_bool = y_with_inf.replace(np.inf, np.nan).isna().values

    # Make KNN model and fitting, prediction
    knn = KNeighborsRegressor(n_neighbors=30)
    knn.fit(X[~na_cell_bool, :], y_with_inf[~na_cell_bool])
    y_filled = y_with_inf.copy()
    y_filled[na_cell_bool] = knn.predict(X[na_cell_bool, :])

    # check
    assert(y_filled.replace(np.inf, np.nan).isna().sum() == 0)

    return y_filled



# -*- coding: utf-8 -*-



import io
import logging
import os
import sys

import pandas as pd
import numpy as np

import h5py

class Data_strage:
    """
    Custom class for quick save and load.
    Attributes can be easily and quickly saved and loaded with hdf5 file.
    Supported attribute types: [int, float, str, list(of int, float, str), np.ndarray (or int, float, str), pd.Dataframe]

    1. If an attribute is stored as an unsupperted type, it will be ignored.
    2. If an attribute name starts with "_", it will be ignored.
    """

    def __init__(self):
        self._exemptions_when_del_attrs = []

    def _set_hdf_path(self, path, key, create_if_not_exist=True):

        if not path.endswith(".hdf5"):
            raise ValueError("path should ends with '.hdf5'")

        self._path = path
        self._key = key
        self._names = []


        try:
            with h5py.File(self._path, mode='r') as f:
                f.visit(self._names.append)

        except:
            if create_if_not_exist:
                print("No hdf file found in the path. New hdf5 file was created.")
                with h5py.File(self._path, mode='w') as f:
                    pass

    def _save_attrs(self, place, attributes):
        """
        Args:
            place (str): Directry in hdf5 file.
            attributes (list of str): attributes to save.
        """

        with h5py.File(self._path, mode='r+') as f:
            for i, j in enumerate(attributes):
                name_ = f"{place}/{j}"
                if name_ in self._names:
                    del f[name_]
                att = getattr(self, j)
                try:
                    f[name_] = att
                except:
                    print(j)
                    f[name_] = att.astype(h5py.string_dtype(encoding='utf-8'))
                self._names.append(name_)
        self._names = list(set(self._names))

    def _load_attrs(self, place, attributes):
        with h5py.File(self._path, mode='r') as f:
            for i, j in enumerate(attributes):
                val = f[f"{place}/{j}"][...]
                setattr(self, j, val)

    def _save_attrs_list(self, place, attributes):
        with h5py.File(self._path, mode='r+') as f:
            for i, j in enumerate(attributes):
                name_ = f"{place}/{j}"
                if name_ in self._names:
                    del f[name_]
                att = getattr(self, j)
                if len(att) == 0:
                    f[name_] = np.array(att)
                elif isinstance(att[0], str):
                    f[name_] = np.array(att).astype(h5py.string_dtype(encoding='utf-8'))
                    self._names.append(name_)
                elif type(att[0]) in [int, float]:
                    f[name_] = np.array(att)
                    self._names.append(name_)
        self._names = list(set(self._names))

    def _load_attrs_list(self, place, attributes):
        with h5py.File(self._path, mode='r') as f:
            for i, j in enumerate(attributes):
                val = f[f"{place}/{j}"][...]
                if len(val) > 0:
                    if isinstance(val[0], bytes):
                        setattr(self, j, [i.decode() for i in val])
                    else:
                        setattr(self, j, list(val))
                else:
                    setattr(self, j, [])

    def _save_attrs_misc(self, place, attributes):
        with h5py.File(self._path, mode='r+') as f:
            for i, j in enumerate(attributes):
                name_ = f"{place}/{j}"
                if name_ in self._names:
                    del f[name_]
                att = getattr(self, j)
                if isinstance(att, str):
                    f[name_] = np.array([att]).astype(h5py.string_dtype(encoding='utf-8'))
                    self._names.append(name_)
                elif type(att) in [int, float]:
                    f[name_] = np.array([att])
                    self._names.append(name_)
        self._names = list(set(self._names))

    def _load_attrs_misc(self, place, attributes):
        with h5py.File(self._path, mode='r') as f:
            for i, j in enumerate(attributes):
                val = f[f"{place}/{j}"][...]
                setattr(self, j, val[0])

    def _save_attrs_df(self, place, attributes):
        """
        Args:
            place (str): Directry in hdf5 file.
            attributes (list of str): attributes to save.
        """
        for i, j in enumerate(attributes):
            key = f"{place}/{j}"
            att = getattr(self, j)
            att.to_hdf(self._path, key=key)

    def _load_attrs_df(self, place, attributes):
        """
        Args:
            place (str): Directry in hdf5 file.
            attributes (list of str): attributes to load.
        """
        for i, j in enumerate(attributes):
            key = f"{place}/{j}"
            att = pd.read_hdf(self._path, key=key)
            setattr(self, j, att)

    def _load_attrs_None(self, place, attributes):

        for i, j in enumerate(attributes):
            setattr(self, j, None)

    def _dump_hdf5(self, place=None):
        """
        Save all attribues in hdf5 file.

        Args:
            place (str): The key for saving data into hdf5 file.

        """
        if place is None:
            place = self._key


        attrs_df = []
        attrs_np = []
        attrs_list = []
        attrs_misc = []
        attrs_None = []

        for key, val in self.__dict__.items():
            if not key.startswith("_"):
                if type(val) in [pd.core.frame.DataFrame]:
                    attrs_df.append(key)
                elif type(val) in [np.ndarray]:
                    attrs_np.append(key)
                elif type(val) in [list]:
                    attrs_list.append(key)
                elif type(val) in [int, str, float]:
                    attrs_misc.append(key)
                elif val is None:
                    attrs_None.append(key)

        self.attrs_df = attrs_df
        self.attrs_np = attrs_np
        self.attrs_list = attrs_list
        self.attrs_misc = attrs_misc
        self.attrs_None = attrs_None

        self._save_attrs_df(place, attrs_df)
        self._save_attrs_list(place, attrs_list + ["attrs_df", "attrs_np", "attrs_list", "attrs_misc", "attrs_None"])
        self._save_attrs_misc(place, attrs_misc)
        self._save_attrs(place, attrs_np)


    def _load_hdf5(self, place=None, specify_attributes=None):
        """
        Load all attributes that was saved in hdf5 file.

        If you enter the list of attributes for specify_attributes, you can select the attribute to be loaded.
        This feature might be useful if you don't waste time and memory to load whold attributes.

        Args:
            place (str): The key in the hdf5 file that was used when save attributes.
            specify_attributes (list): Attribute names to be loaded.

        """

        if place is None:
            place = self._key


        self._load_attrs_list(place, ["attrs_df", "attrs_np", "attrs_list", "attrs_misc", "attrs_None"])

        if specify_attributes is not None:
            attrs_np = [i for i in self.attrs_np if i in specify_attributes]
            attrs_misc = [i for i in self.attrs_misc if i in specify_attributes]
            attrs_list = [i for i in self.attrs_list if i in specify_attributes]
            attrs_df = [i for i in self.attrs_df if i in specify_attributes]
            attrs_None = [i for i in self.attrs_None if i in specify_attributes]
        else:
            attrs_np = self.attrs_np
            attrs_misc = self.attrs_misc
            attrs_list = self.attrs_list
            attrs_df = self.attrs_df
            attrs_None = self.attrs_None

        self._load_attrs(place, attrs_np)
        self._load_attrs_misc(place, attrs_misc)
        self._load_attrs_list(place, attrs_list)
        self._load_attrs_df(place, attrs_df)
        self._load_attrs_None(place, attrs_None)

    def _del_attrs(self, exemptions=[]):
        """
        Delete attributes in the object.
        """

        del_exemptions = self._exemptions_when_del_attrs + exemptions
        for key, val in self.__dict__.items():
            if (not key.startswith("_")) & (not key in del_exemptions):
                setattr(self, key, None)



import pandas as pd
import numpy as np
import matplotlib.pyplot as plt


import seaborn as sns


import os, sys, shutil, importlib, glob
from tqdm.notebook import tqdm

from celloracle import motif_analysis as ma
import celloracle as co
print("co version: ", co.__version__)

plt.rcParams['figure.figsize'] = [6, 4.5]
plt.rcParams["savefig.dpi"] = 300

def main():
    test_tutorial_process_peak_data()

def test_tutorial_process_peak_data():
    # Start analysis
    import time
    start = time.time()
    
    # 1. Load data
    print("1. Load data")
    # 1.0 Download demo data
    print()
    from celloracle.utility.data_download_from_web import download_demo_data
    download_demo_data(file="all_peaks.csv")
    download_demo_data(file="cicero_connections.csv")

    # 1.1. Load data
    # Load scATAC-seq peak list.
    peaks = pd.read_csv("all_peaks.csv", index_col=0)
    peaks = peaks.x.values
    print(peaks)

    # Load Cicero coaccessibility scores.
    cicero_connections = pd.read_csv("cicero_connections.csv", index_col=0)
    cicero_connections.head()

    # 2. Annotate TSS
    print("2. Annotate TSS")
    print(ma.SUPPORTED_REF_GENOME)

    ##!! Please make sure to specify the correct reference genome here
    tss_annotated = ma.get_tss_info(peak_str_list=peaks, ref_genome="mm10")
    # Check results
    print(tss_annotated.tail())

    # 3. Integrate TSS info and cicero connections
    print("3. Integrate TSS info and cicero connections")
    integrated = ma.integrate_tss_peak_with_cicero(tss_peak=tss_annotated,
                                               cicero_connections=cicero_connections)
    print(integrated.shape)
    print(integrated.head())

    # 4. Filter peaks
    print("4. Filter peaks")
    peak = integrated[integrated.coaccess >= 0.8]
    peak = peak[["peak_id", "gene_short_name"]].reset_index(drop=True)

    print(peak.shape)
    print(peak.head())

    # 5. Save data
    print("5. Save data")
    peak.to_csv("processed_peak_file.csv")

    # 6. Remove files
    print("6. Remove files")
    for i in ["all_peaks.csv", "cicero_connections.csv", "processed_peak_file.csv"]:
        os.remove(i)
        #pass
    elapsed = time.time() - start

    print("Success")
    print(f"Total time: {elapsed/60:.3g} minutes")


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
"""
The :mod:`.motif_analysis` module implements transcription factor motif scan.

Genomic activity information (peak of ATAC-seq or Chip-seq) is extracted first.
Then the peak DNA sequence will be subjected to TF motif scan.
Finally we will get list of TFs that potentially binds to a specific gene.

"""

from .motif_analysis_utility import is_genome_installed
from .process_bed_file import peak2fasta, read_bed, remove_zero_seq
from .tfinfo_core import (load_TFinfo, load_TFinfo_from_parquets,
                          make_TFinfo_from_scanned_file, TFinfo, scan_dna_for_motifs)
from .reference_genomes import SUPPORTED_REF_GENOME
from .tss_annotation import get_tss_info
from .process_cicero_data import integrate_tss_peak_with_cicero
from . import process_cicero_data
from .motif_data import load_motifs, MOTIFS_LIST#, PATH_TO_MOTIF_DATA, PFM_PATHS, MOTIFS_PATH_DICT, 


__all__ = ["is_genome_installed", "peak2fasta", "read_bed", "remove_zero_seq",
           "load_TFinfo", "load_TFinfo_from_parquets",
           "make_TFinfo_from_scanned_file",
           "TFinfo", "scan_dna_for_motifs", "SUPPORTED_REF_GENOME",
           "get_tss_info", "process_cicero_data",
           "integrate_tss_peak_with_cicero",
           "load_motifs", "PATH_TO_MOTIF_DATA", "PFM_PATHS", "MOTIFS_PATH_DICT", "MOTIFS_LIST"]




import numpy as np

from tqdm.auto import tqdm

#from network_fitting import TransNet as tn
from numba import jit, f8, i8, void


#############################################
### custom function for markov simulation ###
#############################################


#@jit(i8[:,:](i8[:], f8[:,:], i8))
def _walk(start_cell_id_array, transition_prob, n_steps):
    """
    Do markov simulation based on given transition probability.
    Transition probability suppose to be NxN 2d numpy array that
    stands for transition prob between N states.

    Args:
        start_cell_id_array (1d numpy array (int)): position for initiation

        transition_prob (2d numpy array (int)): transition probabiliry matrix

        n_steps (int) : numbers of steps

    Returns:
        2d numpy array: trajectory of idices saved in a matrix.
    """
    n_cells = transition_prob.shape[0]

    # define starting cell index.
    ids_now = start_cell_id_array
    li = []

    li.append(list(ids_now))
    ids_unique = list(range(n_cells))

    # walk for n_steps
    for i in range(n_steps):
        # calculate next position for all cells
        li_ = []
        for j in ids_now:
            choiced = ids_unique[np.searchsorted(np.cumsum(transition_prob[j,:]), np.random.random(), side="right")]
            #choiced = np.random.choice(a=ids_unique, replace=True, p=transition_prob[j,:])
            li_.append(choiced)
        # update current position
        ids_now = np.array(li_)
        li.append(list(ids_now))

    #li = tuple(li)
    trajectory_matrix = np.array(li).transpose()
    return trajectory_matrix



@jit(nopython=True)
def numba_random_seed(value: int) -> None:
    """Same as np.random.seed but for numba"""
    np.random.seed(value)



# -*- coding: utf-8 -*-


import pandas as pd
import numpy as np

import sys
import os
import shutil
import glob

import scanpy as sc

from ..utility import exec_process

rscript_folder = os.path.abspath(os.path.dirname(__file__))


def seurat_object_to_anndata(file_path_seurat_object, delete_tmp_file=True):
    """
    Convert seurat object into anndata.

    Args:
        file_path_seurat_object (str): File path of seurat object. Seurat object should be saved as Rds format.
        delete_tmp_file (bool): Whether to delete temporary file.
    Returns:
        anndata: anndata object.

    """
    # check file name
    print("input file name: " + file_path_seurat_object)
    if not file_path_seurat_object.lower().endswith(".rds"):
        raise ValueError("Seurat object should be saved as .Rds file")

    # run R script to extract information and make mtx files
    #os.makedirs("tmp", exist_ok=True)
    command = f"Rscript {rscript_folder}/seurat_to_mtx.R {file_path_seurat_object}"


    exec_process(command, message=True, wait_finished=True, return_process=False)

    print("making AnnData ...")


    folder = "./tmp"
    assay_files = glob.glob(folder + "/assay_*")
    assay_names = np.unique([os.path.basename(i).split("_")[1] for i in assay_files])

    if len(assay_names) > 1:
        print(f"{len(assay_names)} assays found in the seurat object.")
        print(f"Data is exported as multiple files.")
        print("If the seurat object was made by integrating multiple files, please be careful about 'Simpson's effect' in the inferred GRN.")
        print("Go to CellOracle web documentation for detailed information about this issue.")

    results = {}
    for assay_name in assay_names:
        results[assay_name] = make_anndata_from_files(assay_name=assay_name,
                                                      folder=folder)
    # Delete temporary files
    if delete_tmp_file:
        shutil.rmtree(folder)

    return results


def make_anndata_from_files(assay_name, folder="./tmp"):

    # Load gene expression matrix
    mm = sc.read_mtx(folder + f"/assay_{assay_name}_data.mtx")
    cell_ids = pd.read_csv(folder + f"/assay_{assay_name}_cells.csv").x.values
    gene_ids = pd.read_csv(folder + f"/assay_{assay_name}_genes.csv").x.values

    # Load meta data
    meta_data = pd.read_csv(os.path.join(folder, "meta_data.csv"),index_col=0)
    meta_dtype = pd.read_csv(os.path.join(folder, "meta_data_dtype.csv"),index_col=0)
    categorical_in_meta = meta_dtype[meta_dtype["dtype"] == "factor"].index.values
    for i in meta_data.columns:
        if i in categorical_in_meta:
            meta_data[i] = meta_data[i].astype(np.object) # change dtype

    # Make anndata
    adata = sc.AnnData(mm.X,
                       obs=meta_data.reindex(cell_ids),
                       var=pd.DataFrame(index=gene_ids))

    # Add raw count data if available
    if f"assay_{assay_name}_rawdata.mtx" in os.listdir(folder):
        raw_data = sc.read_mtx(folder + f"/assay_{assay_name}_rawdata.mtx")
        adata.layers["raw_count"] = raw_data.X.copy()


    # Add dimensional reduction data into anndata
    dr_files = glob.glob(folder + "/reduction_*")
    for i in dr_files:
        # Load reduction data
        dr = pd.read_csv(i, index_col=0)
        reduction_name = os.path.basename(i)[10:-4]
        if reduction_name == "FItSNE":
            reduction_name = "tsne"
        adata.obsm[f"X_{reduction_name}"] = dr.values

    # Add variable gene info
    if "var_genes.csv" in os.listdir(folder):
        variable_genes = pd.read_csv(os.path.join(folder, "var_genes.csv")).x.values
        adata.var["variable_gene"] = adata.var.index.isin(variable_genes)

    # Add color data
    color_df = pd.read_csv(os.path.join(folder, "cluster_color_hex.csv"), index_col=0)
    for i in adata.obs.columns:
        if all(adata.obs["active_ident"].values.astype("object") == \
               adata.obs[i].values.astype("object")):
            adata.uns[f"{i}_colors"] = color_df.colors_hex.values

    return adata


def main():
    if len(sys.argv) == 3:
        output_file_name = sys.argv[2]
    else:
        output_file_name = sys.argv[1]

    if not output_file_name.endswith(".h5ad"):
         output_file_name += ".h5ad"

    #print(sys.argv[:])
    adata_dictionary = seurat_object_to_anndata(file_path_seurat_object=sys.argv[1],
                                                delete_tmp_file=True)

    # save
    print("saving AnnData as h5ad file ...")
    if len(adata_dictionary.keys()) >= 2:
        for assay_name, adata in adata_dictionary.items():
            adata.write(filename=output_file_name.replace(".h5ad", f"_{assay_name}.h5ad"))
    else:
        for assay_name, adata in adata_dictionary.items():
            adata.write(filename=output_file_name)
    print("finished")
if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
"""
Produces simple Sankey Diagrams with matplotlib.
@author: Anneya Golob & marcomanz & pierre-sassoulas & jorwoods
                      .-.
                 .--.(   ).--.
      <-.  .-.-.(.->          )_  .--.
       `-`(     )-'             `)    )
         (o  o  )                `)`-'
        (      )                ,)
        ( ()  )                 )
         `---"\    ,    ,    ,/`
               `--' `--' `--'
                |  |   |   |
                |  |   |   |
                '  |   '   |
"""

from collections import defaultdict

import matplotlib
matplotlib.use('Agg')
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns


class PySankeyException(Exception):
    pass


class NullsInFrame(PySankeyException):
    pass


class LabelMismatch(PySankeyException):
    pass


def check_data_matches_labels(labels, data, side):
    if len(labels) > 0:
        if isinstance(data, list):
            data = set(data)
        if isinstance(data, pd.Series):
            data = set(data.unique().tolist())
        if isinstance(labels, list):
            labels = set(labels)
        if labels != data:
            msg = "\n"
            if len(labels) <= 20:
                msg = "Labels: " + ",".join(labels) + "\n"
            if len(data) < 20:
                msg += "Data: " + ",".join(data)
            raise LabelMismatch('{0} labels and data do not match.{1}'.format(side, msg))


def sankey(left, right, leftWeight=None, rightWeight=None, colorDict=None,
           leftLabels=None, rightLabels=None, aspect=4, rightColor=False,
           fontsize=14, figureName=None, closePlot=False):
    '''
    Make Sankey Diagram showing flow from left-->right

    Inputs:
        left = NumPy array of object labels on the left of the diagram
        right = NumPy array of corresponding labels on the right of the diagram
            len(right) == len(left)
        leftWeight = NumPy array of weights for each strip starting from the
            left of the diagram, if not specified 1 is assigned
        rightWeight = NumPy array of weights for each strip starting from the
            right of the diagram, if not specified the corresponding leftWeight
            is assigned
        colorDict = Dictionary of colors to use for each label
            {'label':'color'}
        leftLabels = order of the left labels in the diagram
        rightLabels = order of the right labels in the diagram
        aspect = vertical extent of the diagram in units of horizontal extent
        rightColor = If true, each strip in the diagram will be be colored
                    according to its left label
    Ouput:
        None
    '''
    if leftWeight is None:
        leftWeight = []
    if rightWeight is None:
        rightWeight = []
    if leftLabels is None:
        leftLabels = []
    if rightLabels is None:
        rightLabels = []
    # Check weights
    if len(leftWeight) == 0:
        leftWeight = np.ones(len(left))

    if len(rightWeight) == 0:
        rightWeight = leftWeight

    # Create Dataframe
    if isinstance(left, pd.Series):
        left = left.reset_index(drop=True)
    if isinstance(right, pd.Series):
        right = right.reset_index(drop=True)
    dataFrame = pd.DataFrame({'left': left, 'right': right, 'leftWeight': leftWeight,
                              'rightWeight': rightWeight}, index=range(len(left)))

    if len(dataFrame[(dataFrame.left.isnull()) | (dataFrame.right.isnull())]):
        raise NullsInFrame('Sankey graph does not support null values.')

    # Identify all labels that appear 'left' or 'right'
    allLabels = pd.Series(np.r_[dataFrame.left.unique(), dataFrame.right.unique()]).unique()

    # Identify left labels
    if len(leftLabels) == 0:
        leftLabels = pd.Series(dataFrame.left.unique()).unique()
    else:
        check_data_matches_labels(leftLabels, dataFrame['left'], 'left')

    # Identify right labels
    if len(rightLabels) == 0:
        rightLabels = pd.Series(dataFrame.right.unique()).unique()
    else:
        check_data_matches_labels(rightLabels, dataFrame['right'], 'right')
    # If no colorDict given, make one
    if colorDict is None:
        colorDict = {}
        palette = "hls"
        colorPalette = sns.color_palette(palette, len(allLabels))
        for i, label in enumerate(allLabels):
            colorDict[label] = colorPalette[i]
    else:
        missing = [label for label in allLabels if label not in colorDict.keys()]
        if missing:
            msg = "The colorDict parameter is missing values for the following labels : "
            msg += '{}'.format(', '.join(missing))
            raise ValueError(msg)

    # Determine widths of individual strips
    ns_l = defaultdict()
    ns_r = defaultdict()
    for leftLabel in leftLabels:
        leftDict = {}
        rightDict = {}
        for rightLabel in rightLabels:
            leftDict[rightLabel] = dataFrame[(dataFrame.left == leftLabel) & (dataFrame.right == rightLabel)].leftWeight.sum()
            rightDict[rightLabel] = dataFrame[(dataFrame.left == leftLabel) & (dataFrame.right == rightLabel)].rightWeight.sum()
        ns_l[leftLabel] = leftDict
        ns_r[leftLabel] = rightDict

    # Determine positions of left label patches and total widths
    leftWidths = defaultdict()
    for i, leftLabel in enumerate(leftLabels):
        myD = {}
        myD['left'] = dataFrame[dataFrame.left == leftLabel].leftWeight.sum()
        if i == 0:
            myD['bottom'] = 0
            myD['top'] = myD['left']
        else:
            myD['bottom'] = leftWidths[leftLabels[i - 1]]['top'] + 0.02 * dataFrame.leftWeight.sum()
            myD['top'] = myD['bottom'] + myD['left']
            topEdge = myD['top']
        leftWidths[leftLabel] = myD

    # Determine positions of right label patches and total widths
    rightWidths = defaultdict()
    for i, rightLabel in enumerate(rightLabels):
        myD = {}
        myD['right'] = dataFrame[dataFrame.right == rightLabel].rightWeight.sum()
        if i == 0:
            myD['bottom'] = 0
            myD['top'] = myD['right']
        else:
            myD['bottom'] = rightWidths[rightLabels[i - 1]]['top'] + 0.02 * dataFrame.rightWeight.sum()
            myD['top'] = myD['bottom'] + myD['right']
            topEdge = myD['top']
        rightWidths[rightLabel] = myD

    # Total vertical extent of diagram
    xMax = topEdge / aspect

    # Draw vertical bars on left and right of each  label's section & print label
    for leftLabel in leftLabels:
        plt.fill_between(
            [-0.02 * xMax, 0],
            2 * [leftWidths[leftLabel]['bottom']],
            2 * [leftWidths[leftLabel]['bottom'] + leftWidths[leftLabel]['left']],
            color=colorDict[leftLabel],
            alpha=0.99
        )
        plt.text(
            -0.05 * xMax,
            leftWidths[leftLabel]['bottom'] + 0.5 * leftWidths[leftLabel]['left'],
            leftLabel,
            {'ha': 'right', 'va': 'center'},
            fontsize=fontsize
        )
    for rightLabel in rightLabels:
        plt.fill_between(
            [xMax, 1.02 * xMax], 2 * [rightWidths[rightLabel]['bottom']],
            2 * [rightWidths[rightLabel]['bottom'] + rightWidths[rightLabel]['right']],
            color=colorDict[rightLabel],
            alpha=0.99
        )
        plt.text(
            1.05 * xMax,
            rightWidths[rightLabel]['bottom'] + 0.5 * rightWidths[rightLabel]['right'],
            rightLabel,
            {'ha': 'left', 'va': 'center'},
            fontsize=fontsize
        )

    # Plot strips
    for leftLabel in leftLabels:
        for rightLabel in rightLabels:
            labelColor = leftLabel
            if rightColor:
                labelColor = rightLabel
            if len(dataFrame[(dataFrame.left == leftLabel) & (dataFrame.right == rightLabel)]) > 0:
                # Create array of y values for each strip, half at left value,
                # half at right, convolve
                ys_d = np.array(50 * [leftWidths[leftLabel]['bottom']] + 50 * [rightWidths[rightLabel]['bottom']])
                ys_d = np.convolve(ys_d, 0.05 * np.ones(20), mode='valid')
                ys_d = np.convolve(ys_d, 0.05 * np.ones(20), mode='valid')
                ys_u = np.array(50 * [leftWidths[leftLabel]['bottom'] + ns_l[leftLabel][rightLabel]] + 50 * [rightWidths[rightLabel]['bottom'] + ns_r[leftLabel][rightLabel]])
                ys_u = np.convolve(ys_u, 0.05 * np.ones(20), mode='valid')
                ys_u = np.convolve(ys_u, 0.05 * np.ones(20), mode='valid')

                # Update bottom edges at each label so next strip starts at the right place
                leftWidths[leftLabel]['bottom'] += ns_l[leftLabel][rightLabel]
                rightWidths[rightLabel]['bottom'] += ns_r[leftLabel][rightLabel]
                plt.fill_between(
                    np.linspace(0, xMax, len(ys_d)), ys_d, ys_u, alpha=0.65,
                    color=colorDict[labelColor]
                )
    plt.gca().axis('off')




# -*- coding: utf-8 -*-



import io
import logging
import os
import pickle
import subprocess
import sys

import pandas as pd
import numpy as np
from tqdm.notebook import tqdm
import matplotlib.pyplot as plt
#import h5py

from ..trajectory.oracle_core import Oracle
from .scatter_to_grid import scatter_value_to_grid_value


def subset_oracle_for_development_analysiis(oracle_object, cluster_column_name, cluster):
    """
    Make a subset of oracle object by specifying of cluster.
    This function pick up some of attributes that needed for development analysis rather than whole attributes.

    """

    cells_of_interest = oracle_object.adata.obs[oracle_object.adata.obs[cluster_column_name] == cluster].index.values

    # Create new oracle object and transfer data
    oracle_ = Oracle()
    for i in ["embedding", "delta_embedding", "delta_embedding_random", "corrcoef_random", "adata"]:
        setattr(oracle_, i, getattr(oracle_object, i))

    index_use = np.where(oracle_.adata.obs.index.isin(cells_of_interest))[0]
    for i in ["embedding", "delta_embedding", "delta_embedding_random", "corrcoef_random"]:
        setattr(oracle_, i, getattr(oracle_, i)[index_use])
    oracle_.adata = oracle_.adata[cells_of_interest, :]

    return oracle_

from scipy.stats import wilcoxon

def get_stat_for_inner_product(oracle_object, n_bins=10):

    # Prepare data
    inner_product_stats = pd.DataFrame({"score": oracle_object.inner_product[~oracle_object.mass_filter],
                                        "pseudotime": oracle_object.new_pseudotime[~oracle_object.mass_filter]})


    bins = _get_bins(inner_product_stats.pseudotime, n_bins)
    inner_product_stats["pseudotime_id"] = np.digitize(inner_product_stats.pseudotime, bins) - 1

    try:
        inner_product_stats["stage"] = oracle_object.stage_grid[~oracle_object.mass_filter]

    except:
        print("stage_grid not calculated")


    # stat test
    ps = []
    for i in np.sort(inner_product_stats["pseudotime_id"].unique()):
        pseudotime_ = inner_product_stats[inner_product_stats["pseudotime_id"]==i].score.values
        stat, p = wilcoxon(x=pseudotime_, alternative="less")
        #print(i, p)
        ps.append(p)

    inner_product_stats_grouped = \
        pd.DataFrame({"ip_score_median":  inner_product_stats.groupby("pseudotime_id").median()["score"].values,
                      "ip_score_mean":  inner_product_stats.groupby("pseudotime_id").mean()["score"],
                      "p-value_negative":  ps},
                      index=np.sort(inner_product_stats["pseudotime_id"].unique()))

    return inner_product_stats, inner_product_stats_grouped


def _get_bins(array, n_bins):
    min_ = array.min()
    max_ = array.max()
    width = (max_ - min_)/(n_bins-1)
    return np.arange(min_, max_ + width, width)


class Oracle_development_module():
    def __init__(self):
        pass

    def extract_data_from_oracle(self, oracle_object, min_mass):

        self.oracle_dev = Oracle()
        ## 1. Extract perturb simulation results as grid matrix
        # 1.1.grid matrix and embedding
        for i in ["flow_grid", "flow", "flow_norm_rndm", "embedding"]:
            setattr(self.oracle_dev, i, getattr(oracle_object, i))

        # 1.2. mass_filter for grid matrix
        self.oracle_dev.mass_filter = (oracle_object.total_p_mass < min_mass)

        ## 2. Extract pseudotime data
        self.oracle_dev.pseudotime = oracle_object.adata.obs["pseudotime"].values

        try:
            self.oracle_dev.stage = np.array(oracle_object.adata.obs["Stage"].values)
        except:
            print("Stage not in data")




    def transfer_data_into_grid(self, args={}):

        if not args:
            args = {"method": "knn",
                    "n_knn": 30}

        self.oracle_dev.new_pseudotime = scatter_value_to_grid_value(embedding=self.oracle_dev.embedding,
                                                          grid=self.oracle_dev.flow_grid,
                                                          value=self.oracle_dev.pseudotime,
                                                          **args)
        try:
            self.oracle_dev.stage_grid = scatter_value_to_grid_value(embedding=self.oracle_dev.embedding,
                                                     grid=self.oracle_dev.flow_grid,
                                                     value=self.oracle_dev.stage,
                                                     **{"method": "knn_class",
                                                     "n_knn": 30})
        except:
            print("Stage not in data")

    def calculate_gradient_and_inner_product(self, scale_factor="l2_norm_mean", normalization=None):

        # Gradient calculation
        gradient = get_gradient(value_on_grid=self.oracle_dev.new_pseudotime.copy())

        if normalization == "sqrt":
            gradient = normalize_gradient(gradient, method="sqrt")

        if scale_factor == "l2_norm_mean":
            # divide gradient by the mean of l2 norm.
            l2_norm = np.linalg.norm(gradient, ord=2, axis=1)
            scale_factor = 1 / l2_norm.mean()

        self.oracle_dev.gradient = gradient * scale_factor

        # Calculate inner product between the pseudotime-gradient and the perturb-gradient
        self.oracle_dev.inner_product = np.array([np.dot(i, j) for i, j in zip(self.oracle_dev.flow, self.oracle_dev.gradient)])


    def calculate_stats(self, n_bins=10):
        inner_product_stats, inner_product_stats_grouped, = \
            get_stat_for_inner_product(self.oracle_dev, n_bins)

        self.oracle_dev.inner_product_stats = inner_product_stats
        self.oracle_dev.inner_product_stats_grouped = inner_product_stats_grouped



class Gradient_based_trajecory():
    def __init__(self, adata=None, obsm_key=None, pseudotime_key="pseudotime", cluster_column_name=None, cluster=None, gt=None):

        if adata is not None:
            self.load_adata(adata=adata, obsm_key=obsm_key,
            pseudotime_key=pseudotime_key,cluster_column_name=cluster_column_name,
            cluster=cluster)
        elif gt is not None:
            self.embedding = gt.embedding_whole.copy()
            self.embedding_whole = gt.embedding_whole.copy()
            self.mass_filter = gt.mass_filter_whole.copy()
            self.mass_filter_whole = gt.mass_filter_whole.copy()
            self.gridpoints_coordinates = gt.gridpoints_coordinates.copy()
            self.pseudotime = gt.pseudotime_whole.copy()

    def load_adata(self, adata, obsm_key, pseudotime_key, cluster_column_name=None, cluster=None):

        self.embedding = adata.obsm[obsm_key]
        self.pseudotime = adata.obs[pseudotime_key].values
        self.embedding_whole = self.embedding.copy()
        self.pseudotime_whole = self.pseudotime.copy()

        if (cluster_column_name is not None) & (cluster is not None):
            cells_ix = np.where(adata.obs[cluster_column_name] == cluster)[0]
            self.embedding = self.embedding[cells_ix, :]
            self.pseudotime = self.pseudotime[cells_ix]





    def calculate_mass_filter(self, min_mass=0.01, smooth=0.8, steps=(40, 40), n_neighbors=200, n_jobs=4):

        x_min, y_min = self.embedding_whole.min(axis=0)
        x_max, y_max = self.embedding_whole.max(axis=0)
        xylim = ((x_min, x_max), (y_min, y_max))

        total_p_mass, gridpoints_coordinates = calculate_p_mass(self.embedding, smooth=smooth, steps=steps,
                                  n_neighbors=n_neighbors, n_jobs=n_jobs, xylim=xylim)

        total_p_mass_whole, _ = calculate_p_mass(self.embedding_whole, smooth=smooth, steps=steps,
                                  n_neighbors=n_neighbors, n_jobs=n_jobs, xylim=xylim)

        self.total_p_mass = total_p_mass
        self.mass_filter = (total_p_mass < min_mass)
        self.mass_filter_whole = (total_p_mass_whole < min_mass)
        self.gridpoints_coordinates = gridpoints_coordinates

    def transfer_data_into_grid(self, args={}):

        if not args:
            args = {"method": "knn",
                    "n_knn": 30}

        self.pseudotime_on_grid = scatter_value_to_grid_value(embedding=self.embedding,
                                                          grid=self.gridpoints_coordinates,
                                                          value=self.pseudotime,
                                                          **args)
    def calculate_gradient(self, scale_factor=60, normalization=None):

        # Gradient calculation
        gradient = get_gradient(value_on_grid=self.pseudotime_on_grid.copy())

        if normalization == "sqrt":
            gradient = normalize_gradient(gradient, method="sqrt")

        if scale_factor == "l2_norm_mean":
            # divide gradient by the mean of l2 norm.
            l2_norm = np.linalg.norm(gradient, ord=2, axis=1)
            scale_factor = 1 / l2_norm.mean()

        self.gradient = gradient * scale_factor

    def visualize_dev_flow(self, scale_for_pseudotime=30, s=10, s_grid=30):
        visualize_dev_flow(self, scale_for_pseudotime=scale_for_pseudotime, s=s, s_grid=s_grid)

def aggregate_GT_object(list_GT_object, base_gt=None):

    pseudotime_stack = [i.pseudotime_on_grid for i in list_GT_object]
    gradient_stack = [i.gradient for i in list_GT_object]
    mass_filter_stack = [i.mass_filter for i in list_GT_object]

    new_pseudotime, new_gradient, new_mass_filter = _aggregate_gradients(pseudotime_stack=pseudotime_stack,
                                                    gradient_stack=gradient_stack,
                                                    mass_filter_stack=mass_filter_stack)

    if base_gt is None:
        gt = Gradient_based_trajecory(gt=list_GT_object[0])
        gt.pseudotime_on_grid = new_pseudotime
        gt.gradient = new_gradient

    else:
        gt = base_gt
        gt.pseudotime_on_grid[~new_mass_filter] = new_pseudotime[~new_mass_filter]
        gt.gradient[~new_mass_filter, :] = new_gradient[~new_mass_filter, :]

    return gt

def _aggregate_gradients(pseudotime_stack, gradient_stack, mass_filter_stack):

    new_pseudotime = np.zeros_like(pseudotime_stack[0])
    new_pseudotime_count = np.zeros_like(pseudotime_stack[0])
    new_gradient = np.zeros_like(gradient_stack[0])
    gradient_count = np.zeros_like(gradient_stack[0])
    for fil, pt, gra in zip(mass_filter_stack, pseudotime_stack, gradient_stack):
        new_pseudotime[~fil] += pt[~fil]
        new_pseudotime_count[~fil] +=1
        new_gradient[~fil, :] += gra[~fil, :]
        gradient_count[~fil, :] += 1

    new_pseudotime[new_pseudotime_count != 0] /= new_pseudotime_count[new_pseudotime_count != 0]
    new_gradient[gradient_count != 0] /= gradient_count[gradient_count != 0]
    new_mass_filter = (gradient_count.sum(axis=1) == 0)

    return new_pseudotime, new_gradient, new_mass_filter


def normalize_gradient(gradient, method="sqrt"):
    """
    Normalize length of 2D vector
    """

    if method == "sqrt":

        size = np.sqrt(np.power(gradient, 2).sum(axis=1))
        size_sq = np.sqrt(size)
        size_sq[size_sq == 0] = 1
        factor = np.repeat(np.expand_dims(size_sq, axis=1), 2, axis=1)

    return gradient / factor

from scipy.stats import norm as normal
from sklearn.neighbors import NearestNeighbors


def calculate_p_mass(embedding, smooth=0.5, steps=(40, 40),
                          n_neighbors=100, n_jobs=4, xylim=((None, None), (None, None))):
    """Calculate the velocity using a points on a regular grid and a gaussian kernel

    Note: the function should work also for n-dimensional grid

    Arguments
    ---------
    embedding:

    smooth: float, smooth=0.5
        Higher value correspond to taking in consideration further points
        the standard deviation of the gaussian kernel is smooth * stepsize
    steps: tuple, default
        the number of steps in the grid for each axis
    n_neighbors:
        number of neighbors to use in the calculation, bigger number should not change too much the results..
        ...as soon as smooth is small
        Higher value correspond to slower execution time
    n_jobs:
        number of processes for parallel computing
    xymin:
        ((xmin, xmax), (ymin, ymax))

    Returns
    -------
    total_p_mass: np.ndarray
        density at each point of the grid

    """

    # Prepare the grid
    grs = []
    for dim_i in range(embedding.shape[1]):
        m, M = np.min(embedding[:, dim_i]), np.max(embedding[:, dim_i])

        if xylim[dim_i][0] is not None:
            m = xylim[dim_i][0]
        if xylim[dim_i][1] is not None:
            M = xylim[dim_i][1]

        m = m - 0.025 * np.abs(M - m)
        M = M + 0.025 * np.abs(M - m)
        gr = np.linspace(m, M, steps[dim_i])
        grs.append(gr)

    meshes_tuple = np.meshgrid(*grs)
    gridpoints_coordinates = np.vstack([i.flat for i in meshes_tuple]).T

    nn = NearestNeighbors(n_neighbors=n_neighbors, n_jobs=n_jobs)
    nn.fit(embedding)
    dists, neighs = nn.kneighbors(gridpoints_coordinates)

    std = np.mean([(g[1] - g[0]) for g in grs])
    # isotropic gaussian kernel
    gaussian_w = normal.pdf(loc=0, scale=smooth * std, x=dists)
    total_p_mass = gaussian_w.sum(1)
    gridpoints_coordinates

    return total_p_mass, gridpoints_coordinates

def get_gradient(value_on_grid):
    # Gradient calculation
    n = int(np.sqrt(value_on_grid.shape[0]))
    value_on_grid_as_matrix = value_on_grid.reshape(n, n)
    dy, dx = np.gradient(value_on_grid_as_matrix)
    gradient = np.stack([dx.flatten(), dy.flatten()], axis=1)

    return gradient


def visualize_dev_flow(self, scale_for_pseudotime=30, s=10, s_grid=30):

    embedding_whole = self.embedding_whole
    embedding_of_interest= self.embedding
    mass_filter = self.mass_filter
    mass_filter_whole = self.mass_filter_whole
    gridpoints_coordinates=self.gridpoints_coordinates

    pseudotime_raw = self.pseudotime
    pseudotime_on_grid=self.pseudotime_on_grid

    gradient_pseudotime=self.gradient


    fig, ax = plt.subplots(1, 5, figsize=[25,5])

    ##
    ax_ = ax[0]
    ax_.scatter(embedding_whole[:, 0], embedding_whole[:, 1], c="lightgray", s=s)
    ax_.scatter(embedding_of_interest[:, 0], embedding_of_interest[:, 1], c=pseudotime_raw, cmap="rainbow", s=s)
    ax_.set_title("Pseudotime")
    ax_.axis("off")

    ####
    ax_ = ax[1]
    ax_.scatter(gridpoints_coordinates[mass_filter, 0], gridpoints_coordinates[mass_filter, 1], s=0)
    ax_.scatter(gridpoints_coordinates[~mass_filter_whole, 0], gridpoints_coordinates[~mass_filter_whole, 1],
     c="lightgray", s=s_grid)
    ax_.scatter(gridpoints_coordinates[~mass_filter, 0], gridpoints_coordinates[~mass_filter, 1],
     c=pseudotime_on_grid[~mass_filter], cmap="rainbow", s=s_grid)
    ax_.set_title("Pseudotime on grid")
    ax_.axis("off")



    ###
    ax_ = ax[2]
    #ax_.scatter(gridpoints_coordinates[mass_filter, 0], gridpoints_coordinates[mass_filter, 1], s=0)
    ax_.scatter(gridpoints_coordinates[mass_filter, 0], gridpoints_coordinates[mass_filter, 1], s=0)
    ax_.scatter(gridpoints_coordinates[~mass_filter_whole, 0], gridpoints_coordinates[~mass_filter_whole, 1],
     c="lightgray", s=s_grid)
    ax_.scatter(gridpoints_coordinates[~mass_filter, 0], gridpoints_coordinates[~mass_filter, 1],
     c=pseudotime_on_grid[~mass_filter], cmap="rainbow", s=s_grid)

    ax_.quiver(gridpoints_coordinates[~mass_filter, 0], gridpoints_coordinates[~mass_filter, 1],
                    gradient_pseudotime[~mass_filter, 0], gradient_pseudotime[~mass_filter, 1],
                    scale=scale_for_pseudotime)
    ax_.set_title("Gradient of pseudotime \n(=Development flow)")
    ax_.axis("off")

    ###
    ax_ = ax[3]
    #ax_.scatter(gridpoints_coordinates[mass_filter, 0], gridpoints_coordinates[mass_filter, 1], s=0)
    ax_.scatter(embedding_whole[:, 0], embedding_whole[:, 1], c="lightgray", s=s)
    ax_.quiver(gridpoints_coordinates[~mass_filter, 0], gridpoints_coordinates[~mass_filter, 1],
                    gradient_pseudotime[~mass_filter, 0], gradient_pseudotime[~mass_filter, 1],
                    scale=scale_for_pseudotime)
    ax_.set_title("Gradient of pseudotime \n(=Development flow)")
    ax_.axis("off")

    ####
    ax_ = ax[4]
    ax_.scatter(embedding_whole[:, 0], embedding_whole[:, 1], c="lightgray", s=s)
    ax_.scatter(embedding_of_interest[:, 0], embedding_of_interest[:, 1], c=pseudotime_raw, cmap="rainbow", s=s)

    ax_.quiver(gridpoints_coordinates[~mass_filter, 0], gridpoints_coordinates[~mass_filter, 1],
                    gradient_pseudotime[~mass_filter, 0], gradient_pseudotime[~mass_filter, 1],
                    scale=scale_for_pseudotime)
    ax_.set_title("Pseudotime + \nDevelopment flow")
    ax_.axis("off")



# -*- coding: utf-8 -*-
import pandas as pd
from scanpy import AnnData

import celloracle as co

#
def test_load_drosophila_promoter_base_GRN_dl():
    df = co.data.load_drosophila_promoter_base_GRN(force_download=True)
    assert isinstance(df, pd.core.frame.DataFrame)

def test_load_drosophila_promoter_base_GRN():
    df = co.data.load_drosophila_promoter_base_GRN(force_download=False)
    assert isinstance(df, pd.core.frame.DataFrame)

#
def test_load_rat_promoter_base_GRN_dl():
    df = co.data.load_rat_promoter_base_GRN(force_download=True)
    assert isinstance(df, pd.core.frame.DataFrame)

def test_load_rat_promoter_base_GRN():
    df = co.data.load_rat_promoter_base_GRN(force_download=False)
    assert isinstance(df, pd.core.frame.DataFrame)

#
def test_load_mouse_promoter_base_GRN_dl():
    df = co.data.load_mouse_promoter_base_GRN(force_download=True)
    assert isinstance(df, pd.core.frame.DataFrame)

def test_load_mouse_promoter_base_GRN():
    df = co.data.load_mouse_promoter_base_GRN(force_download=False)
    assert isinstance(df, pd.core.frame.DataFrame)

#
def test_load_human_promoter_base_GRN_dl():
    df = co.data.load_human_promoter_base_GRN(force_download=True)
    assert isinstance(df, pd.core.frame.DataFrame)

def test_load_human_promoter_base_GRN():
    df = co.data.load_human_promoter_base_GRN(force_download=False)
    assert isinstance(df, pd.core.frame.DataFrame)

#
def test_load_chicken_promoter_base_GRN_dl():
    df = co.data.load_chicken_promoter_base_GRN(force_download=True)
    assert isinstance(df, pd.core.frame.DataFrame)

def test_load_chicken_promoter_base_GRN():
    df = co.data.load_chicken_promoter_base_GRN(force_download=False)
    assert isinstance(df, pd.core.frame.DataFrame)

#
def test_load_xenopus_tropicalis_promoter_base_GRN_dl():
    df = co.data.load_xenopus_tropicalis_promoter_base_GRN(force_download=True)
    assert isinstance(df, pd.core.frame.DataFrame)

def test_load_xenopus_tropicalis_promoter_base_GRN():
    df = co.data.load_xenopus_tropicalis_promoter_base_GRN(force_download=False)
    assert isinstance(df, pd.core.frame.DataFrame)

#
def test_load_xenopus_laevis_promoter_base_GRN_dl():
    df = co.data.load_xenopus_laevis_promoter_base_GRN(force_download=True)
    assert isinstance(df, pd.core.frame.DataFrame)

def test_load_xenopus_laevis_promoter_base_GRN():
    df = co.data.load_xenopus_laevis_promoter_base_GRN(force_download=False)
    assert isinstance(df, pd.core.frame.DataFrame)

#
def test_load_arabidopsis_promoter_base_GRN_dl():
    df = co.data.load_arabidopsis_promoter_base_GRN(force_download=True)
    assert isinstance(df, pd.core.frame.DataFrame)

def test_load_arabidopsis_promoter_base_GRN():
    df = co.data.load_arabidopsis_promoter_base_GRN(force_download=False)
    assert isinstance(df, pd.core.frame.DataFrame)

#
def test_load_Scerevisiae_promoter_base_GRN_dl():
    df = co.data.load_Scerevisiae_promoter_base_GRN(force_download=True)
    assert isinstance(df, pd.core.frame.DataFrame)

def test_load_Scerevisiae_promoter_base_GRN():
    df = co.data.load_Scerevisiae_promoter_base_GRN(force_download=False)
    assert isinstance(df, pd.core.frame.DataFrame)

#
def test_load_Celegans_promoter_base_GRN_dl():
    df = co.data.load_Celegans_promoter_base_GRN(force_download=True)
    assert isinstance(df, pd.core.frame.DataFrame)

def test_load_Celegans_promoter_base_GRN():
    df = co.data.load_Celegans_promoter_base_GRN(force_download=False)
    assert isinstance(df, pd.core.frame.DataFrame)

#
def test_load_Pig_promoter_base_GRN_dl():
    df = co.data.load_Pig_promoter_base_GRN(force_download=True)
    assert isinstance(df, pd.core.frame.DataFrame)

def test_load_Pig_promoter_base_GRN():
    df = co.data.load_Pig_promoter_base_GRN(force_download=False)
    assert isinstance(df, pd.core.frame.DataFrame)
